class Drawable:
    """A base class for all classes holding opengl data."""

    def draw(self, *args, **kwargs):
        """Draw the opengl data."""
        raise NotImplementedError

    def unload(self):
        """Unload the opengl data."""
        raise NotImplementedError



"""A collection of classes for wx objects to abstract away
the repeated code and make working with wx a bit more simple."""

import logging
import wx
from wx.lib.scrolledpanel import ScrolledPanel
from typing import Iterable, Union, Any, List, Optional, Sequence, Dict, Tuple

log = logging.getLogger(__name__)


class SimpleSizer:
    def __init__(self, sizer_dir=wx.VERTICAL):
        self._sizer = self.sizer = wx.BoxSizer(sizer_dir)

    def add_object(self, obj, space=1, options=wx.ALL):
        self.sizer.Add(obj, space, options, 5)


class SimplePanel(wx.Panel, SimpleSizer):
    def __init__(self, parent: wx.Window, sizer_dir=wx.VERTICAL):
        wx.Panel.__init__(
            self,
            parent,
            wx.ID_ANY,
            wx.DefaultPosition,
            wx.DefaultSize,
            wx.TAB_TRAVERSAL,
        )
        SimpleSizer.__init__(self, sizer_dir)
        self.SetSizer(self.sizer)


class SimpleScrollablePanel(ScrolledPanel, SimpleSizer):
    """A scrolled panel that automatically sets itself up."""

    def __init__(self, parent: wx.Window, sizer_dir=wx.VERTICAL, **kwargs):
        ScrolledPanel.__init__(self, parent, **kwargs)
        SimpleSizer.__init__(self, sizer_dir)
        self.SetSizer(self.sizer)
        self.SetupScrolling()
        self.SetAutoLayout(1)

    def DoGetBestSize(self):
        sizer = self.GetSizer()
        if sizer is None:
            return -1, -1
        else:
            sx, sy = sizer.CalcMin()
            return (
                sx + wx.SystemSettings.GetMetric(wx.SYS_VSCROLL_X),
                sy,
            )


class SimpleChoice(wx.Choice):
    """A wrapper for wx.Choice that sets up the UI for you."""

    def __init__(
        self,
        parent: wx.Window,
        choices: Sequence[str] = (),
        default: Optional[str] = None,
    ):
        super().__init__(parent, choices=choices)
        if choices:
            if default is not None and default in choices:
                self.SetSelection(choices.index(default))
            else:
                self.SetSelection(0)

    def GetCurrentString(self) -> str:
        return self.GetString(self.GetSelection())


StringableType = Any


class SimpleChoiceAny(wx.Choice):
    """An extension for wx.Choice that enables showing and returning objects that are not strings."""

    def __init__(self, parent: wx.Window, sort=True, reverse=False):
        super().__init__(parent)
        self._values: List[Any] = []  # the data hidden behind the string
        self._keys: List[str] = []  # the strings shown to the user
        self._sorted = sort
        self._reverse = reverse

    @property
    def keys(self) -> Tuple[str, ...]:
        """Get the string values displayed to the user"""
        return tuple(self._keys)

    @property
    def values(self) -> Tuple[Any, ...]:
        """Get the data hidden behind the string value"""
        return tuple(self._values)

    @property
    def items(self) -> Tuple[Tuple[str, Any], ...]:
        """Get the string value and the data hidden behind the value"""
        return tuple(zip(self._keys, self._values))

    def SetItems(
        self,
        items: Union[Iterable[StringableType], Dict[StringableType, Any]],
        default: StringableType = None,
    ):
        """Set items. Does not have to be strings.
        If items is a dictionary the string of the values are show to the user and the key is returned from GetAny
        If it is just an iterable the string of the values are shown and the raw equivalent input is returned.
        """
        if not items:
            return
        if isinstance(items, dict):
            items: List[Tuple[str, Any]] = [
                (str(value), key) for key, value in items.items()
            ]
            if self._sorted:
                items = sorted(items, key=lambda x: x[0], reverse=self._reverse)
            self._keys = [key.strip() for key, _ in items]
            self._values = [value for _, value in items]
        else:
            if self._sorted:
                self._values = list(sorted(items))
                if self._reverse:
                    self._values.reverse()
            else:
                self._values = list(items)
            self._keys = [str(v).strip() for v in self._values]
        super().SetItems(self._keys)
        if default is not None and default in self._values:
            self.SetSelection(self._values.index(default))
        else:
            self.SetSelection(0)

    def SetValue(self, value: Any):
        if value in self._keys:
            self.SetSelection(self._keys.index(value))

    def GetAny(self) -> Optional[Any]:
        """Return the value currently selected in the form before it was converted to a string"""
        log.warning(
            "SimpleChoiceAny.GetAny is being depreciated and will be removed in the future. Please use SimpleChoiceAny.GetCurrentObject instead",
            exc_info=True,
        )
        return self.GetCurrentObject()

    def GetCurrentObject(self) -> Optional[Any]:
        """Return the value currently selected in the form before it was converted to a string"""
        if self._values:
            return self._values[self.GetSelection()]

    def GetCurrentString(self) -> str:
        """Get the string form of the value currently selected."""
        return self.GetString(self.GetSelection())


class SimpleDialog(wx.Dialog):
    """A dialog with ok and cancel buttons set up."""

    def __init__(self, parent: wx.Window, title, sizer_dir=wx.VERTICAL):
        wx.Dialog.__init__(
            self, parent, title=title, style=wx.CAPTION | wx.RESIZE_BORDER
        )
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)
        self.sizer = wx.BoxSizer(sizer_dir)
        sizer.Add(self.sizer, 1, wx.EXPAND)
        self.bottom_sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(self.bottom_sizer, 0, wx.EXPAND)
        self.bottom_sizer.AddStretchSpacer()
        button_sizer = self.CreateButtonSizer(wx.OK | wx.CANCEL)
        self.bottom_sizer.Add(button_sizer, flag=wx.ALL, border=5)



from typing import List, Optional, Tuple, Any
import numpy
from OpenGL.GL import (
    glCullFace,
    GL_FRONT,
    GL_BACK,
    glGetIntegerv,
    GL_CULL_FACE_MODE,
)

from amulet.api.level import BaseLevel
from amulet.api.data_types import FloatTriplet, PointCoordinates, Dimension

from amulet_map_editor.api.opengl import (
    Drawable,
    ContextManager,
    ThreadedObjectContainer,
)
from amulet_map_editor.api.opengl.mesh.level import RenderLevel
from amulet_map_editor.api.opengl.matrix import (
    transform_matrix,
    displacement_matrix,
)
from amulet_map_editor.api.opengl.resource_pack import (
    OpenGLResourcePackManager,
    OpenGLResourcePack,
)

LocationType = PointCoordinates
ScaleType = FloatTriplet
RotationType = FloatTriplet
TransformType = Tuple[LocationType, ScaleType, RotationType]


class LevelGroup(
    OpenGLResourcePackManager, Drawable, ThreadedObjectContainer, ContextManager
):
    """A group of RenderLevel classes with transforms"""

    def __init__(
        self,
        context_identifier: Any,
        resource_pack: OpenGLResourcePack,
    ):
        OpenGLResourcePackManager.__init__(self, resource_pack)
        ThreadedObjectContainer.__init__(self)
        ContextManager.__init__(self, context_identifier)
        self._objects: List[RenderLevel] = []
        self._transforms: List[TransformType] = []
        self._world_translation: List[LocationType] = []
        self._transformation_matrices: List[numpy.ndarray] = []
        self._is_mirrored: List[bool] = []
        self._active_level_index: Optional[int] = None
        self._camera_location: LocationType = (0.0, 100.0, 0.0)

    @property
    def active_level_index(self) -> Optional[int]:
        """The index of the active level. None if no level is active."""
        return self._active_level_index

    @property
    def render_levels(self) -> Tuple[RenderLevel, ...]:
        return tuple(self._objects)

    @property
    def active_transform(self) -> TransformType:
        """Get the transform of the active level.
        If no level is selected will return zeros."""
        if self._active_level_index is None:
            return (0, 0, 0), (0, 0, 0), (0, 0, 0)
        return self._transforms[self._active_level_index]

    @active_transform.setter
    def active_transform(self, location_scale_rotation: TransformType):
        """Set the transform for the active object.
        Has no effect if there is no active object.
        :param location_scale_rotation: The location, scale and rotation
        :return:
        """
        if self._active_level_index is not None:
            location, scale, rotation = location_scale_rotation
            self._transforms[self._active_level_index] = (location, scale, rotation)
            self._is_mirrored[self._active_level_index] = bool(
                sum(1 for s in scale if s < 0) % 2
            )
            self._transformation_matrices[self._active_level_index] = numpy.matmul(
                transform_matrix(scale, rotation, location),
                displacement_matrix(*self._world_translation[self._active_level_index]),
            )
            self._set_camera_location()

    def set_camera_location(self, x: float, y: float, z: float):
        """Set the location of the camera for each of the levels."""
        self._camera_location = (x, y, z)
        self._set_camera_location()

    def _set_camera_location(self):
        for level, transform in zip(self._objects, self._transformation_matrices):
            try:
                level.camera_location = numpy.matmul(
                    numpy.linalg.inv(transform), (*self._camera_location, 1)
                ).tolist()[:-1]
            except numpy.linalg.LinAlgError:
                pass

    def set_camera_rotation(self, yaw: float, pitch: float):
        """Set the rotation of the camera for each of the levels."""
        for level in self._objects:
            level.camera_rotation = yaw, pitch

    def append(
        self,
        level: BaseLevel,
        dimension: Dimension,
        location: LocationType,
        scale: ScaleType,
        rotation: RotationType,
    ):
        """Append a level to the list and activate it."""
        # TODO: update this to support multiple levels
        self.clear()
        render_level = RenderLevel(
            self.context_identifier,
            self._resource_pack,
            level,
            draw_box=True,
            limit_bounds=True,
        )
        render_level.dimension = dimension
        # the level objects to be drawn
        self.register(render_level)
        # the transforms (tuple) applied by the user
        self._transforms.append((location, scale, rotation))
        self._is_mirrored.append(bool(sum(1 for s in scale if s < 0) % 2))
        self._world_translation.append(
            (
                -(
                    (
                        level.bounds(dimension).min_array
                        + level.bounds(dimension).max_array
                    )
                    // 2
                ).astype(int)
            ).tolist()
        )
        # the matrix of the transform applied by the user
        self._transformation_matrices.append(
            numpy.matmul(
                transform_matrix(scale, rotation, location),
                displacement_matrix(*self._world_translation[-1]),
            )
        )
        if self._active_level_index is None:
            self._active_level_index = 0
        else:
            self._active_level_index += 1

    def enable(self):
        """Enable chunk generation in a new thread."""
        for level in self._objects:
            level.enable()

    def unload(self):
        """Unload the geometry. Frees VRAM."""
        for level in self._objects:
            level.unload()

    def clear(self):
        """Destroy and unload all level objects."""
        self.unload()
        for level in self._objects.copy():
            self.unregister(level)
        self._transforms.clear()
        self._is_mirrored.clear()
        self._world_translation.clear()
        self._transformation_matrices.clear()
        self._active_level_index = None

    def run_garbage_collector(self):
        for level in self._objects:
            level.run_garbage_collector()

    def _rebuild(self):
        self.unload()

    def draw(self, camera_matrix: numpy.ndarray):
        """Draw all of the levels."""
        for level, transform, is_mirrored in zip(
            self._objects, self._transformation_matrices, self._is_mirrored
        ):
            cull_state = glGetIntegerv(GL_CULL_FACE_MODE)
            if is_mirrored:
                glCullFace(GL_FRONT)
            else:
                glCullFace(GL_BACK)
            level.draw(numpy.matmul(camera_matrix, transform))
            glCullFace(cull_state)



from typing import Optional
import wx
from amulet_map_editor.api import image


class TracebackDialog(wx.Dialog):
    def __init__(
        self,
        parent: Optional[wx.Window],
        title: str,
        error: str,
        traceback: str,
        **kwargs,
    ):
        kwargs["style"] = (
            kwargs.get("style", 0) | wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER
        )
        wx.Dialog.__init__(self, parent, **kwargs)
        self.SetTitle(title)

        self._error = error
        self._traceback = traceback

        main_sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(main_sizer)

        error_sizer = wx.BoxSizer(wx.HORIZONTAL)
        main_sizer.Add(error_sizer, 0, wx.EXPAND | wx.ALL, 5)

        bitmap_1 = wx.StaticBitmap(
            self, bitmap=image.icon.tablericons.alert_circle.bitmap(32, 32)
        )
        error_sizer.Add(bitmap_1, 0, 0, 0)

        error_text = wx.StaticText(self, wx.ID_ANY, error)
        error_sizer.Add(error_text, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)

        traceback_text = wx.TextCtrl(
            self,
            wx.ID_ANY,
            traceback,
            style=wx.TE_MULTILINE | wx.TE_READONLY,
            size=(800, -1),
        )
        main_sizer.Add(traceback_text, 1, wx.EXPAND, 0)

        button_sizer = wx.StdDialogButtonSizer()
        main_sizer.Add(button_sizer, 0, wx.ALIGN_RIGHT | wx.ALL, 4)

        copy_button = wx.Button(self, wx.ID_ANY, "Copy Error")
        copy_button.Bind(wx.EVT_BUTTON, self._on_copy_error)
        button_sizer.Add(copy_button, 0, 0, 0)

        button_ok = wx.Button(self, wx.ID_OK, "")
        button_ok.SetDefault()
        button_sizer.AddButton(button_ok)

        button_sizer.Realize()

        main_sizer.Fit(self)

        self.SetAffirmativeId(button_ok.GetId())

        self.Fit()
        self.Layout()

    def _on_copy_error(self, evt):
        if wx.TheClipboard.Open():
            wx.TheClipboard.SetData(
                wx.TextDataObject(f"{self._error}\n{self._traceback}")
            )
            wx.TheClipboard.Close()


class MyApp(wx.App):
    def OnInit(self):
        import traceback as traceback_

        try:
            raise Exception("Exception message")
        except Exception as e:
            err = str(e)
            tb = traceback_.format_exc()
        self.dialog = TracebackDialog(None, "title", err, tb)
        self.SetTopWindow(self.dialog)
        self.dialog.ShowModal()
        self.dialog.Destroy()
        return True


if __name__ == "__main__":
    app = MyApp(0)
    app.MainLoop()



from typing import Tuple

import PyMCTranslate
import wx

from amulet_map_editor.api.wx.ui.base_define import BaseDefine
from amulet_map_editor.api.wx.ui.biome_select.biome_select import BiomeSelect


class BiomeDefine(BaseDefine):
    def __init__(
        self,
        parent,
        translation_manager: PyMCTranslate.TranslationManager,
        orientation=wx.VERTICAL,
        platform: str = None,
        version_number: Tuple[int, int, int] = None,
        namespace: str = None,
        biome_name: str = None,
        show_pick_biome: bool = False,
        **kwargs,
    ):
        super().__init__(
            parent,
            translation_manager,
            BiomeSelect,
            orientation,
            platform,
            version_number,
            namespace,
            default_name=biome_name,
            show_pick=show_pick_biome,
            show_force_blockstate=False,
            **kwargs,
        )

    def _on_picker_change(self, evt):
        evt.Skip()

    @property
    def biome_name(self) -> str:
        return self._picker.name

    @biome_name.setter
    def biome_name(self, biome_name: str):
        self._picker.name = biome_name

    @property
    def biome(self) -> str:
        return f"{self.namespace}:{self.biome_name}"

    @biome.setter
    def biome(self, biome: str):
        namespace, biome_name = biome.split(":")
        self._picker.set_namespace(namespace)
        self._picker.set_name(biome_name)

    @property
    def universal_biome(self) -> str:
        return self._translation_manager.get_version(
            self.platform, self.version_number
        ).biome.to_universal(self.biome)

    @universal_biome.setter
    def universal_biome(self, universal_biome: str):
        self.biome = self._translation_manager.get_version(
            self.platform, self.version_number
        ).biome.from_universal(universal_biome)



from amulet_map_editor.programs.edit.api.ui.tool import BaseOperationChoiceToolUI


class OperationTool(BaseOperationChoiceToolUI):
    OperationGroupName = "operations"

    @property
    def name(self) -> str:
        return "Operation"



from typing import List


class ThreadedObject:
    """Data/Geometry generation controlled by an external thread."""

    def thread_action(self):
        """The action the thread will call."""
        raise NotImplementedError

    @property
    def thread_weighting(self) -> int:
        """The number of times the thread_action should be called."""
        return 1


class ThreadedObjectContainer(ThreadedObject):
    def __init__(self):
        self._objects: List[ThreadedObject] = []
        self._obj_index = 0
        self._obj_sub_index = 0

    def register(self, thread_object: ThreadedObject):
        assert isinstance(thread_object, ThreadedObject)
        if thread_object in self._objects:
            raise Exception("ThreadedObject object is already registered.")
        self._objects.append(thread_object)

    def unregister(self, thread_object: ThreadedObject):
        assert isinstance(thread_object, ThreadedObject)
        if thread_object not in self._objects:
            raise Exception("ThreadedObject object was not already registered.")
        self._objects.remove(thread_object)

    def thread_action(self):
        """The action the thread will call."""
        if self._obj_index >= len(self._objects):
            self._obj_index = 0
            self._obj_sub_index = 0
        while (
            self._obj_index < len(self._objects)
            and self._objects[self._obj_index].thread_weighting < self._obj_sub_index
        ):
            self._obj_index += 1
            self._obj_sub_index = 0

        if self._obj_index < len(self._objects):
            self._objects[self._obj_index].thread_action()
            self._obj_sub_index += 1

    @property
    def thread_weighting(self) -> int:
        """The number of times the thread_action should be called."""
        return sum(obj.thread_weighting for obj in self._objects)



import logging
from OpenGL.GL import (
    glBindTexture,
    GL_TEXTURE_2D,
    GL_TRIANGLES,
    glBindVertexArray,
    glBindBuffer,
    GL_ARRAY_BUFFER,
    GL_STATIC_DRAW,
    glUseProgram,
    glGetUniformLocation,
    glGenVertexArrays,
    glGenBuffers,
    glVertexAttribPointer,
    GL_FLOAT,
    GL_FALSE,
    glEnableVertexAttribArray,
    glBufferData,
    glDeleteBuffers,
    glDeleteVertexArrays,
    glUniformMatrix4fv,
    glUniform1i,
    glActiveTexture,
    GL_TEXTURE0,
    glDrawArrays,
)
from OpenGL.error import GLError
import ctypes
import numpy
from amulet_map_editor.api.opengl.shaders import get_shader
from amulet_map_editor.api.opengl import Drawable, ContextManager
from amulet_map_editor.api.opengl.resource_pack import (
    OpenGLResourcePackManagerStatic,
    OpenGLResourcePack,
)

log = logging.getLogger(__name__)


class TriMesh(Drawable, OpenGLResourcePackManagerStatic, ContextManager):
    """The base class for a triangular face mesh.
    Implements the base logic to set up and unload OpenGL."""

    _vertex_attrs = (
        3,  # vertex attribute pointers
        2,  # texture coords attribute pointers
        4,  # texture coords attribute pointers
        3,  # tint value (also shading)
    )
    _vert_len = sum(_vertex_attrs)

    def __init__(self, context_identifier: str, resource_pack: OpenGLResourcePack):
        """Create a new TriMesh.
        The object can be created from another thread so OpenGL
        variables cannot be set from here"""
        OpenGLResourcePackManagerStatic.__init__(self, resource_pack)
        ContextManager.__init__(self, context_identifier)
        self._vao = None  # vertex array object
        self._vbo = None  # vertex buffer object
        self._shader = None  # the shader program
        self._transform_location = (
            None  # the reference within the shader program of the transformation matrix
        )
        self._texture_location = None  # the location of the texture in the shader
        self._texture = None
        self.verts = self.new_empty_verts()  # the vertices to draw
        self.draw_start = 0
        self.draw_count = 0  # the number of vertices to draw

    @staticmethod
    def new_empty_verts() -> numpy.ndarray:
        return numpy.zeros(0, dtype=numpy.float32)

    @property
    def vertex_usage(self):
        return GL_STATIC_DRAW

    @property
    def draw_mode(self):
        return GL_TRIANGLES

    @property
    def shader_name(self) -> str:
        return "render_chunk"

    def _setup(self):
        """Setup OpenGL attributes if required"""
        if self._vao is None:  # if the opengl state has not been set
            self._texture = self.resource_pack.get_atlas_id(self.context_identifier)
            self._shader = get_shader(self.context_identifier, self.shader_name)
            glUseProgram(self._shader)
            self._transform_location = glGetUniformLocation(
                self._shader, "transformation_matrix"
            )
            self._texture_location = glGetUniformLocation(self._shader, "image")
            self._vao = glGenVertexArrays(1)  # create the array
            glBindVertexArray(self._vao)
            self._vbo = glGenBuffers(1)  # and the buffer
            glBindBuffer(GL_ARRAY_BUFFER, self._vbo)
            self._setup_opengl_attrs()
            self._change_verts()
            glBindVertexArray(0)
            glBindBuffer(GL_ARRAY_BUFFER, 0)
            glUseProgram(0)

    def _setup_opengl_attrs(self):
        """Set up OpenGL vertex attributes"""
        attr_start = 0
        for index, attr_count in enumerate(self._vertex_attrs):
            glVertexAttribPointer(
                index,
                attr_count,
                GL_FLOAT,
                GL_FALSE,
                self._vert_len * 4,
                ctypes.c_void_p(attr_start * 4),
            )
            glEnableVertexAttribArray(index)
            attr_start += attr_count

    def change_verts(self, verts=None):
        """Modify the vertices in OpenGL."""
        log.debug(f"change_verts {self}")
        try:
            glBindVertexArray(self._vao)
            glBindBuffer(GL_ARRAY_BUFFER, self._vbo)
        except GLError:  # There seems to be errors randomly when binding the VBO
            log.debug(
                f"Failed binding the OpenGL state for {self}. Trying to reload it."
            )
            self.unload()
            self._setup()
            if verts is None:
                return
            glBindVertexArray(self._vao)
            glBindBuffer(GL_ARRAY_BUFFER, self._vbo)
        self._change_verts(verts)
        glBindVertexArray(0)
        glBindBuffer(GL_ARRAY_BUFFER, 0)

    def _change_verts(self, verts=None):
        """Modify the vertices in OpenGL. Requires binding and unbinding."""
        if verts is not None:
            glBufferData(GL_ARRAY_BUFFER, verts.size * 4, verts, self.vertex_usage)
        else:
            glBufferData(
                GL_ARRAY_BUFFER, self.verts.size * 4, self.verts, self.vertex_usage
            )

    def unload(self):
        """Unload all opengl data"""
        log.debug(f"unload {self}")
        if self._vbo is not None:
            glDeleteBuffers(1, int(self._vbo))
            self._vbo = None
        if self._vao is not None:
            glDeleteVertexArrays(1, int(self._vao))
            self._vao = None

    def draw(self, transformation_matrix: numpy.ndarray):
        self._setup()
        self._draw(transformation_matrix)

    def _draw(self, transformation_matrix: numpy.ndarray):
        glUseProgram(self._shader)
        glUniformMatrix4fv(
            self._transform_location,
            1,
            GL_FALSE,
            transformation_matrix.T.astype(numpy.float32),
        )
        glUniform1i(self._texture_location, 0)
        try:
            glBindVertexArray(self._vao)
        except GLError:  # There seems to be errors randomly when binding the VBO
            log.debug(
                f"Failed binding the OpenGL state for {self}. Trying to reload it."
            )
            self.unload()
            self._setup()
            glBindVertexArray(self._vao)
        glActiveTexture(GL_TEXTURE0)
        glBindTexture(GL_TEXTURE_2D, self._texture)

        glDrawArrays(self.draw_mode, self.draw_start, self.draw_count)

        glBindVertexArray(0)
        glUseProgram(0)



import pickle
from typing import Any, TYPE_CHECKING, Union, Tuple
import os
import wx
import weakref
import warnings

from amulet_map_editor.programs.edit.api.edit_canvas_container import (
    EditCanvasContainer,
)
from amulet_map_editor.programs.edit.api.ui.canvas_toggle_element import (
    CanvasToggleElement,
)

if TYPE_CHECKING:
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas
    from amulet.api.level import BaseLevel

OperationUIType = Union[wx.Window, wx.Sizer, "OperationUI"]


class OperationUI(EditCanvasContainer, CanvasToggleElement):
    """The base class that all operations must inherit from."""

    def __init__(
        self,
        parent: wx.Window,
        canvas: "EditCanvas",
        world: "BaseLevel",
        options_path: str,
    ):
        super().__init__(canvas)
        self._parent = weakref.ref(parent)
        self._world = weakref.ref(world)
        self._options_path = options_path

    @property
    def parent(self) -> wx.Window:
        """The UI element that contains this operation."""
        return self._parent()

    @property
    def world(self) -> "BaseLevel":
        """The world to operate on."""
        return self._world()

    @property
    def wx_add_options(self) -> Tuple[int, ...]:
        """The options used to put the UI element into a vertical BoxSizer covering the whole canvas.
        Override if these options do not work for your UI."""
        if isinstance(self, wx.Window):
            return ()
        elif isinstance(self, wx.Sizer):
            return 1, wx.EXPAND
        return ()

    def enable(self):
        """Set the state of the UI for being enabled.
        Do not bind events to the canvas here because they will get removed. Do this in bind_events.
        """
        pass

    def bind_events(self):
        """Bind all required events to the canvas.
        All events on the canvas will be automatically removed after the UI is disabled.
        """
        pass

    def disable(self):
        """Stop the UI and make it safe to be destroyed.
        Unload any excessive data. May get resumed again with a call to enable.
        All events bound to the canvas will be automatically removed after this is run.
        """
        if hasattr(self, "unload") and callable(self.unload):
            warnings.warn(
                "OperationUI.unload is depreciated and will be removed in the future. Please used OperationUI.disable instead",
                DeprecationWarning,
            )
            self.unload()

    def _load_options(self, default=None) -> Any:
        """Load previously saved options from disk or return the default options."""
        try:
            with open(self._options_path, "rb") as f:
                return pickle.load(f)
        except:
            return default

    def _save_options(self, options: Any):
        """Save the given options to disk so that they persist in the next session."""
        os.makedirs(os.path.dirname(self._options_path), exist_ok=True)
        with open(self._options_path, "wb") as f:
            return pickle.dump(options, f)



import wx
from wx import glcanvas
from .camera import Camera


_SpeedChangedEventType = wx.NewEventType()
EVT_SPEED_CHANGED = wx.PyEventBinder(_SpeedChangedEventType)


class SpeedChangedEvent(wx.PyEvent):
    """Run when the speed of the camera has changed."""

    def __init__(self, speed: float):
        wx.PyEvent.__init__(self, eventType=_SpeedChangedEventType)
        self._speed = speed

    @property
    def speed(self) -> float:
        """The speed of the camera."""
        return self._speed


class ControllableCamera(Camera):
    """A slight modification to the default camera to add speed attributes."""

    __slots__ = ("_move_speed", "_rotate_speed")

    def __init__(self, canvas: glcanvas.GLCanvas):
        super().__init__(canvas)
        self._move_speed = 2.0
        self._rotate_speed = 2.0
        self._rotating = False

    @property
    def move_speed(self) -> float:
        """The speed that the camera moves at."""
        return self._move_speed

    @move_speed.setter
    def move_speed(self, val: float):
        """Set the speed that the camera moves at."""
        self._move_speed = float(val)
        wx.PostEvent(self.canvas, SpeedChangedEvent(self.move_speed))

    @property
    def rotate_speed(self) -> float:
        """The speed that the camera rotates at."""
        return self._rotate_speed

    @rotate_speed.setter
    def rotate_speed(self, val: float):
        """Set the speed that the camera rotates at."""
        self._rotate_speed = float(val)

    @property
    def rotating(self):
        """Is the camera rotating (True) or fixed (False)"""
        return self._rotating

    @rotating.setter
    def rotating(self, rotating: bool):
        """Set if the camera is rotating (True) or fixed (False)"""
        self._rotating = bool(rotating)
        self._notify_moved()



# Fixed function example plugin 3
# see example 2 before looking at this example
# This example shows how to use options with the fixed function pipeline


from amulet.api.selection import SelectionGroup
from amulet.api.level import BaseLevel
from amulet.api.data_types import Dimension


# function example for the Button input.
def function_name():
    print("you have clicked a button.")


operation_options = {  # options is a dictionary where the key is the description shown to the user and the value describes how to build the UI
    "Text Label": ["label"],  # This will just be a label
    # bool examples  https://wxpython.org/Phoenix/docs/html/wx.CheckBox.html
    "Bool input default": [
        "bool"
    ],  # This will be a check box that takes the default value (unchecked)
    "Bool input False": [
        "bool",
        False,
    ],  # This will be a check box that starts unchecked
    "Bool input True": ["bool", True],  # This will be a check box that starts checked
    # int examples  https://wxpython.org/Phoenix/docs/html/wx.SpinCtrl.html
    "Int input default": ["int"],  # This will be an integer input with starting value 0
    "Int input 10": [
        "int",
        10,
    ],  # This will be an integer input with starting value 10 (or whatever you put in the second slot)
    "Int input 10 bounded": [
        "int",
        10,
        0,
        20,
    ],  # same as above but must be between the third and fourth values
    # float examples  https://wxpython.org/Phoenix/docs/html/wx.SpinCtrlDouble.html
    "Float input default": [
        "float"
    ],  # This will be a float input with starting value 0.0
    "Float input 10": [
        "float",
        10,
    ],  # This will be a float input with starting value 10.0 (or whatever you put in the second slot)
    "Float input 10 bounded": [
        "float",
        10,
        0,
        20,
    ],  # same as above but must be between the third and fourth values
    # string input examples  https://wxpython.org/Phoenix/docs/html/wx.TextCtrl.html
    "String input empty": [
        "str"
    ],  # This will be a text input with an empty starting value
    "String input empty2": ["str", ""],  # Same as the above
    "String input hello": ["str", "hello"],  # Text entry with starting value "hello"
    # string choice examples  https://wxpython.org/Phoenix/docs/html/wx.Choice.html
    "Text choice": ["str_choice", "choice 1", "choice 2", "choice 3"],
    # OS examples
    "File Open picker": ["file_open"],  # UI to pick an existing file, will be blank
    "File Save picker": ["file_save"],  # UI to pick a file to save to, will be blank
    "Folder picker": ["directory"],  # UI to pick a directory, will be blank
    "File Open picker": ["file_open", path],  # UI to pick an existing file with a path
    "File Save picker": ["file_save", path],  # UI to pick a file to save to with a path
    "Folder picker": ["directory", path],  # UI to pick a directory with a path
    # UI to add a button and it will run the function_name() when clicked
    "Button input": ["button", "button name", function_name],
}


def operation(
    world: BaseLevel, dimension: Dimension, selection: SelectionGroup, options: dict
):
    # When the user presses the run button this function will be run as normal but
    # since the "options" key was defined in export this function will get another
    # input in the form of a dictionary where the keys are the same as you defined
    # them in the options dictionary above and the values are what the user picked
    # in the UI (bool, int, float, str)
    # If "options" is not defined in export this will just be an empty dictionary
    pass


export = {
    "name": "Fixed Function Pipeline Example 3",
    "operation": operation,
    "options": operation_options,  # The options you defined above should be added here to show in the UI
}



# cython: language_level=3, boundscheck=False, wraparound=False
# distutils: extra_compile_args = OPENMP_CCARGS
# distutils: extra_link_args = OPENMP_CLARGS

import numpy
cimport numpy
numpy.import_array()

from libc.stdlib cimport malloc, calloc, free

cdef extern from "stdlib.h":
    void *memcpy(void *dest, void *src, size_t n) nogil

from cpython cimport array
import array
from cython.parallel import prange

cdef extern from *:
    """
    #ifdef _OPENMP
    #pragma message ("openmp enabled")
    #else
    #pragma message ("openmp disabled")
    #endif
    """

cdef float _brightness_step = 0.15
cdef dict _brightness_multiplier = {
    None: (1,) * 3,
    "up": (1,) * 3,
    "north": (1 - _brightness_step,) * 3,
    "south": (1 - _brightness_step,) * 3,
    "east": (1 - _brightness_step * 2,) * 3,
    "west": (1 - _brightness_step * 2,) * 3,
    "down": (1 - _brightness_step * 3,) * 3,
}

CULL_STR_INDEX = {
    None: 0,
    "up": 1,
    "down": 2,
    "north": 3,
    "east": 4,
    "south": 5,
    "west": 6,
}

cdef int CULL_MAP[7][3]
CULL_MAP[0][:] = (0, 0, 0)
CULL_MAP[1][:] = (0, 1, 0)
CULL_MAP[2][:] = (0, -1, 0)
CULL_MAP[3][:] = (0, 0, -1)
CULL_MAP[4][:] = (1, 0, 0)
CULL_MAP[5][:] = (0, 0, 1)
CULL_MAP[6][:] = (-1, 0, 0)

DEF ARRAY_VERT_COUNT = 10_008  # The number of vertices in the table
DEF ATTR_COUNT = 12  # The number of float attributes per vertex
DEF ARRAY_SIZE = ARRAY_VERT_COUNT * ATTR_COUNT

cdef struct BlockArray:
    unsigned int* arr  # pointer to the array
    int sx  # the shape of the array in the x axis
    int sy  # the shape of the array in the y axis
    int sz  # the shape of the array in the z axis
    long dx # the displacement in the x axis
    long dy # the displacement in the y axis
    long dz # the displacement in the z axis

cdef BlockArray* BlockArray_new(int sx, int sy, int sz) nogil:
    self = <BlockArray*>calloc(1, sizeof(BlockArray))
    self.arr = <unsigned int*>malloc(sx * sy * sz * sizeof(unsigned int))
    self.sx = sx
    self.sy = sy
    self.sz = sz
    self.dx = 0
    self.dy = 0
    self.dz = 0
    return self

cdef BlockArray* BlockArray_init(
        unsigned int* arr,
        int sx,
        int sy,
        int sz,
        long dx,
        long dy,
        long dz,
) nogil:
    self = BlockArray_new(sx, sy, sz)
    memcpy(self.arr, arr, sx * sy * sz * sizeof(unsigned int))
    self.dx = dx
    self.dy = dy
    self.dz = dz
    return self

cdef void BlockArray_free(BlockArray* self) nogil:
    free(self.arr)
    free(self)


cdef struct VertArray:
    float* arr  # pointer to the array
    int size  # the number of floats in the array

cdef VertArray* VertArray_new(unsigned long size) nogil:
    # assert size and size % (ATTR_COUNT*3) == 0, "arr must have a multiple of 36 values"
    vert_array = <VertArray*>calloc(1, sizeof(VertArray))
    vert_array.arr = <float*>malloc(size * sizeof(float))
    vert_array.size = size
    return vert_array

cdef VertArray* VertArray_init(float* arr, unsigned long size) nogil:
    vert_array = VertArray_new(size)
    memcpy(vert_array.arr, arr, size * sizeof(float))
    return vert_array

cdef void VertArray_free(VertArray* vert_array) nogil:
    free(vert_array.arr)
    free(vert_array)

cdef VertArray* VertArray_from_py(array.array arr):
    assert arr.typecode == "f", "arr must be a float array"
    return VertArray_init(&arr.data.as_floats[0], len(arr))


cdef struct BlockModel:
    VertArray* faces[7]
    char is_transparent

cdef BlockModel* BlockModel_init(dict face_data, char is_transparent):
    block_model = <BlockModel*>calloc(1, sizeof(BlockModel))
    block_model.is_transparent = is_transparent
    cdef Py_ssize_t index
    for cull_id, index in CULL_STR_INDEX.items():
        if cull_id in face_data:
            arr = face_data[cull_id]
            if isinstance(arr, numpy.ndarray):
                arr = array.array("f", arr.ravel())
            if isinstance(arr, array.array) and arr.typecode == "f":
                block_model.faces[index] = VertArray_from_py(arr)
        else:
            block_model.faces[index] = NULL
    return block_model

cdef void BlockModel_free(BlockModel* block_model) nogil:
    cdef Py_ssize_t i
    for i in range(7):
        if block_model.faces[i]:
            VertArray_free(block_model.faces[i])
    free(block_model)


cdef class BlockModelManager:
    cdef BlockModel** blocks  # A pointer to an array of pointers to BlockModel structs
    cdef unsigned long block_size  # The size of the blocks array
    cdef unsigned long block_count  # The amount of the blocks array that is used

    def __cinit__(self):
        self.blocks = NULL
        self.block_size = 0
        self.block_count = 0

    def __init__(self):
        self.blocks = <BlockModel**>calloc(100, sizeof(BlockModel*))
        self.block_size = 100

    def __dealloc__(self):
        for i in range(self.block_count):
            BlockModel_free(self.blocks[i])
        free(self.blocks)

    cdef _extend(self):
        if self.block_count == self.block_size:
            blocks_temp = <BlockModel**>calloc(self.block_size + 100, sizeof(BlockModel*))
            memcpy(blocks_temp, self.blocks, self.block_size * sizeof(BlockModel*))
            free(self.blocks)
            self.blocks = blocks_temp
            self.block_size += 100

    cpdef add_block(self, dict face_data, int is_transparent):
        self._extend()
        self.blocks[self.block_count] = BlockModel_init(face_data, is_transparent)
        self.block_count += 1

    def __len__(self):
        return self.block_count


cdef struct VertArrayContainer:
    VertArray** arrays  # A pointer to an array of pointers to VertArrays
    int size  # The current size of the array
    int used  # The number of elements of the array that are used

cdef VertArrayContainer* VertArrayContainer_init() nogil:
    self = <VertArrayContainer*>calloc(1, sizeof(VertArrayContainer))
    self.arrays = <VertArray**>calloc(10, sizeof(VertArray*))
    self.size = 0
    self.used = 0
    return self

cdef void VertArrayContainer_free(VertArrayContainer* self) nogil:
    cdef int i
    for i in range(self.used):
        VertArray_free(self.arrays[i])
    free(self.arrays)
    free(self)

cdef void VertArrayContainer_append(VertArrayContainer* self, VertArray* vert_array) nogil:
    if self.used == self.size:
        _VertArrayContainer_extend(self)
    self.arrays[self.used] = vert_array
    self.used += 1

cdef void _VertArrayContainer_extend(VertArrayContainer* self) nogil:
    arr = <VertArray**>calloc(self.size + 5, sizeof(VertArray*))
    memcpy(arr, self.arrays, self.size * sizeof(VertArray*))
    free(self.arrays)
    self.arrays = arr
    self.size += 5


cdef struct VertArrayContainerTuple:
    VertArrayContainer* verts
    VertArrayContainer* verts_translucent

cdef VertArrayContainerTuple* VertArrayContainerTuple_init() nogil:
    self = <VertArrayContainerTuple*>calloc(1, sizeof(VertArrayContainerTuple))
    self.verts = VertArrayContainer_init()
    self.verts_translucent = VertArrayContainer_init()
    return self

cdef void VertArrayContainerTuple_free(VertArrayContainerTuple* self) nogil:
    VertArrayContainer_free(self.verts)
    VertArrayContainer_free(self.verts_translucent)
    free(self)

cdef unsigned int get_block(
    BlockArray* block_array,
    int x,
    int y,
    int z
) nogil:
    return block_array.arr[x * block_array.sz * block_array.sy + y * block_array.sz + z]

cdef VertArrayContainerTuple* create_lod0_sub_chunk(
    BlockArray* block_array,
    BlockModelManager block_model_manager,
) nogil:
    cdef int x, y, z, x_, y_, z_, dx, dy, dz  # location variables

    # float counters
    cdef unsigned int vert_count, vert_end, vertex, vertex_attr

    cdef float shade
    cdef unsigned int block_id, cull_id
    cdef BlockModel* block_model
    cdef VertArray* vert_array

    cdef VertArray* vert_table = VertArray_new(ARRAY_SIZE)
    vert_table.size = 0
    cdef VertArray* trans_vert_table = VertArray_new(ARRAY_SIZE)
    trans_vert_table.size = 0

    cdef int size_x = block_array.sx - 2
    cdef int size_y = block_array.sy - 2
    cdef int size_z = block_array.sz - 2

    cdef VertArrayContainerTuple* verts = VertArrayContainerTuple_init()

    for x in range(size_x):
        for y in range(size_y):
            for z in range(size_z):
                x_ = x + 1
                y_ = y + 1
                z_ = z + 1
                block_id = get_block(block_array, x_, y_, z_)
                block_model = block_model_manager.blocks[block_id]
                for cull_id in range(7):
                    if block_model.faces[cull_id]:
                        # iterate through each cull direction
                        if cull_id:
                            dx = x_ + CULL_MAP[cull_id][0]
                            dy = y_ + CULL_MAP[cull_id][1]
                            dz = z_ + CULL_MAP[cull_id][2]
                            # If the next block is opaque or both blocks are full transparent blocks, do nothing
                            if block_model_manager.blocks[get_block(block_array, dx, dy, dz)].is_transparent == 0 or \
                               block_model_manager.blocks[get_block(block_array, dx, dy, dz)].is_transparent == block_model_manager.blocks[get_block(block_array, x_, y_, z_)].is_transparent == 1:
                                continue

                        vert_array = block_model.faces[cull_id]
                        vert_count = vert_array.size

                        if block_model.is_transparent == 1:
                            vert_end = trans_vert_table.size+vert_count
                            if vert_end > ARRAY_SIZE:
                                VertArrayContainer_append(verts.verts_translucent, trans_vert_table)
                                trans_vert_table = VertArray_new(ARRAY_SIZE)
                                trans_vert_table.size = 0
                                vert_end = vert_count
                            memcpy(&trans_vert_table.arr[trans_vert_table.size], vert_array.arr, vert_count * sizeof(float))
                            for vertex in range(trans_vert_table.size, vert_end, ATTR_COUNT):
                                trans_vert_table.arr[vertex + 0] += block_array.dx + x
                                trans_vert_table.arr[vertex + 1] += block_array.dy + y
                                trans_vert_table.arr[vertex + 2] += block_array.dz + z
                                shade = ((trans_vert_table.arr[vertex + 1] / 32) % 2)
                                if shade > 1:
                                    shade = - shade + 2
                                shade = 0.9 + 0.2 * shade
                                for vertex_attr in range(9, 12):
                                    trans_vert_table.arr[vertex + vertex_attr] *= shade
                            trans_vert_table.size += vert_count
                        else:
                            vert_end = vert_table.size+vert_count
                            if vert_end > ARRAY_SIZE:
                                VertArrayContainer_append(verts.verts, vert_table)
                                vert_table = VertArray_new(ARRAY_SIZE)
                                vert_table.size = 0
                                vert_end = vert_count
                            memcpy(&vert_table.arr[vert_table.size], vert_array.arr, vert_count * sizeof(float))
                            for vertex in range(vert_table.size, vert_end, ATTR_COUNT):
                                vert_table.arr[vertex + 0] += block_array.dx + x
                                vert_table.arr[vertex + 1] += block_array.dy + y
                                vert_table.arr[vertex + 2] += block_array.dz + z
                                shade = ((vert_table.arr[vertex + 1] / 32) % 2)
                                if shade > 1:
                                    shade = - shade + 2
                                shade = 0.9 + 0.2 * shade
                                for vertex_attr in range(9, 12):
                                    vert_table.arr[vertex + vertex_attr] *= shade
                            vert_table.size += vert_count

    if vert_table.size:
        VertArrayContainer_append(verts.verts, vert_table)
    else:
        VertArray_free(vert_table)
    if trans_vert_table.size:
        VertArrayContainer_append(verts.verts_translucent, trans_vert_table)
    else:
        VertArray_free(trans_vert_table)

    return verts


cdef tuple _create_lod0_chunk(
    BlockModelManager block_model_manager,
    list blocks,
    long[:] chunk_offset
):
    cdef int i, j
    cdef long sub_chunk_y
    cdef unsigned int[:, :, ::1] block_array
    cdef int sub_chunk_count = len(blocks)
    block_array_list = <BlockArray**>calloc(sub_chunk_count, sizeof(BlockArray*))
    sub_chunk_verts = <VertArrayContainerTuple**>calloc(sub_chunk_count, sizeof(VertArrayContainerTuple*))

    for i in range(sub_chunk_count):
        block_array, sub_chunk_y = blocks[i]
        block_array_list[i] = BlockArray_init(
            &block_array[0, 0, 0],
            block_array.shape[0],
            block_array.shape[1],
            block_array.shape[2],
            chunk_offset[0],
            chunk_offset[1] + sub_chunk_y,
            chunk_offset[2]
        )

    for i in prange(sub_chunk_count, nogil=True):
    # for i in range(sub_chunk_count):
        sub_chunk_verts[i] = create_lod0_sub_chunk(
            block_array_list[i],
            block_model_manager,
        )

    for i in range(sub_chunk_count):
        BlockArray_free(block_array_list[i])
    free(block_array_list)

    cdef unsigned long vert_size = 0
    cdef unsigned long vert_size_translucent = 0
    cdef VertArrayContainerTuple* vert_array_container_tuple
    cdef VertArrayContainer* vert_array_container
    cdef VertArray* vert_array

    for i in range(sub_chunk_count):
        vert_array_container_tuple = sub_chunk_verts[i]
        vert_array_container = vert_array_container_tuple.verts
        for j in range(vert_array_container.used):
            vert_array = vert_array_container.arrays[j]
            vert_size += vert_array.size
        vert_array_container = vert_array_container_tuple.verts_translucent
        for j in range(vert_array_container.used):
            vert_array = vert_array_container.arrays[j]
            vert_size_translucent += vert_array.size

    chunk_verts = array.clone(array.array("f"), vert_size, zero=False)
    chunk_verts_translucent = array.clone(array.array("f"), vert_size_translucent, zero=False)

    vert_size = 0
    vert_size_translucent = 0

    for i in range(sub_chunk_count):
        vert_array_container_tuple = sub_chunk_verts[i]
        vert_array_container = vert_array_container_tuple.verts
        for j in range(vert_array_container.used):
            vert_array = vert_array_container.arrays[j]
            memcpy(&chunk_verts.data.as_floats[vert_size], vert_array.arr, vert_array.size * sizeof(float))
            vert_size += vert_array.size
        vert_array_container = vert_array_container_tuple.verts_translucent
        for j in range(vert_array_container.used):
            vert_array = vert_array_container.arrays[j]
            memcpy(&chunk_verts_translucent.data.as_floats[vert_size_translucent], vert_array.arr, vert_array.size * sizeof(float))
            vert_size_translucent += vert_array.size

    for i in range(sub_chunk_count):
        VertArrayContainerTuple_free(sub_chunk_verts[i])
    free(sub_chunk_verts)

    return [chunk_verts], [chunk_verts_translucent]


def _extend_blocks(resource_pack, block_palette):
    # TODO: This is kind of janky
    #  set up a proper location for these
    #  This needs to be cached somewhere so that the geometry does not need recreating each time.
    #  The issue is that the cache is dependent on the resource pack being used and the block palette it is applied to
    if not hasattr(resource_pack, "block_model_manager"):
        resource_pack.block_model_manager = {}
    if id(block_palette) not in resource_pack.block_model_manager:
        resource_pack.block_model_manager[id(block_palette)] = BlockModelManager()

    block_model_manager = resource_pack.block_model_manager[id(block_palette)]

    done_count = len(block_model_manager)
    state_count = len(block_palette)

    for block_id in range(done_count, state_count):
        # more block states have been added

        model = resource_pack.get_block_model(
            block_palette[block_id]
        )
        vert_map = {}
        for py_cull_dir in model.faces.keys():
            if py_cull_dir in CULL_STR_INDEX:
                # the vertices in model space
                verts = model.verts[py_cull_dir].reshape((-1, 3))
                tverts = model.texture_coords[py_cull_dir].reshape((-1, 2))
                faces = model.faces[py_cull_dir]

                py_vert_table = numpy.zeros(
                    (faces.size, ATTR_COUNT), dtype=numpy.float32
                )
                py_vert_table[:, :3] = verts[faces]
                py_vert_table[:, 3:5] = tverts[faces]

                vert_index = 0
                for texture_index in model.texture_index[py_cull_dir]:
                    py_vert_table[vert_index : vert_index + 3, 5:9] = resource_pack.texture_bounds(model.textures[texture_index])
                    vert_index += 3

                py_vert_table[:, 9:12] = (
                    model.tint_verts[py_cull_dir].reshape((-1, 3))[faces]
                    * _brightness_multiplier[py_cull_dir]
                )
                vert_map[py_cull_dir] = py_vert_table
        block_model_manager.add_block(vert_map, model.is_transparent)


def create_lod0_chunk(
    resource_pack,
    chunk_offset: numpy.ndarray,
    blocks,
    block_palette,
    vert_len,  # should be 12
):
    _extend_blocks(resource_pack, block_palette)
    return _create_lod0_chunk(
        resource_pack.block_model_manager[id(block_palette)],
        blocks,
        chunk_offset
    )



import wx
from typing import TYPE_CHECKING, Type, Dict, Optional

from amulet_map_editor.programs.edit.api import EditCanvasContainer
from amulet_map_editor.programs.edit.api.ui.tool.base_tool_ui import (
    BaseToolUI,
    BaseToolUIType,
)
from amulet_map_editor.programs.edit.api.events import (
    ToolChangeEvent,
    EVT_TOOL_CHANGE,
)

from amulet_map_editor.programs.edit.plugins.tools import (
    ImportTool,
    ExportTool,
    OperationTool,
    SelectTool,
    ChunkTool,
    PasteTool,
)

if TYPE_CHECKING:
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas


class ToolManagerSizer(wx.BoxSizer, EditCanvasContainer):
    def __init__(self, canvas: "EditCanvas"):
        wx.BoxSizer.__init__(self, wx.VERTICAL)
        EditCanvasContainer.__init__(self, canvas)

        self._tools: Dict[str, BaseToolUIType] = {}
        self._active_tool: Optional[BaseToolUIType] = None

        self._tool_option_sizer = wx.BoxSizer(wx.VERTICAL)
        self.Add(
            self._tool_option_sizer, 1, wx.EXPAND | wx.RESERVE_SPACE_EVEN_IF_HIDDEN, 0
        )

        tool_select_sizer = wx.BoxSizer(wx.HORIZONTAL)
        tool_select_sizer.AddStretchSpacer(1)
        self._tool_select = ToolSelect(canvas)
        tool_select_sizer.Add(self._tool_select, 0, wx.EXPAND, 0)
        tool_select_sizer.AddStretchSpacer(1)
        self.Add(tool_select_sizer, 0, wx.EXPAND, 0)

        self.register_tool(SelectTool)
        self.register_tool(PasteTool)
        self.register_tool(OperationTool)
        self.register_tool(ImportTool)
        self.register_tool(ExportTool)
        self.register_tool(ChunkTool)

    @property
    def tools(self):
        return self._tools.copy()

    def bind_events(self):
        if self._active_tool is not None:
            self._active_tool.bind_events()
        self.canvas.Bind(EVT_TOOL_CHANGE, self._enable_tool_event)

    def register_tool(self, tool_cls: Type[BaseToolUIType]):
        assert issubclass(tool_cls, (wx.Window, wx.Sizer)) and issubclass(
            tool_cls, BaseToolUI
        )
        tool = tool_cls(self.canvas)
        self._tool_select.register_tool(tool.name)
        if isinstance(tool, wx.Window):
            tool.Hide()
        elif isinstance(tool, wx.Sizer):
            tool.ShowItems(show=False)
        self._tools[tool.name] = tool
        self._tool_option_sizer.Add(tool, 1, wx.EXPAND, 0)

    def _enable_tool_event(self, evt: ToolChangeEvent):
        self._enable_tool(evt.tool, evt.state)

    def enable(self):
        if isinstance(self._active_tool, SelectTool):
            self._active_tool.enable()
            self.canvas.reset_bound_events()
            self.canvas.Layout()
        else:
            self._enable_tool("Select")

    def disable(self):
        """Disable the active tool."""
        if self._active_tool is not None:
            self._active_tool.disable()

    def enable_default_tool(self):
        """
        Enables the default tool (the select tool)
        """
        if not isinstance(self._active_tool, SelectTool):
            self._enable_tool("Select")

    def _enable_tool(self, tool: str, state=None):
        if tool in self._tools:
            if self._active_tool is not None:
                self._active_tool.disable()
                if isinstance(self._active_tool, wx.Window):
                    self._active_tool.Hide()
                elif isinstance(self._active_tool, wx.Sizer):
                    self._active_tool.ShowItems(show=False)
            self._active_tool = self._tools[tool]
            if isinstance(self._active_tool, wx.Window):
                self._active_tool.Show()
            elif isinstance(self._active_tool, wx.Sizer):
                self._active_tool.ShowItems(show=True)
            self._active_tool.enable()
            self._active_tool.set_state(state)
            self.canvas.reset_bound_events()
            self.canvas.Layout()


class ToolSelect(wx.Panel, EditCanvasContainer):
    def __init__(self, canvas: "EditCanvas"):
        wx.Panel.__init__(self, canvas)
        EditCanvasContainer.__init__(self, canvas)

        self._sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.SetSizer(self._sizer)

    def register_tool(self, name: str):
        button = wx.Button(self, label=name)
        self._sizer.Add(button)
        self._sizer.Fit(self)
        self.Layout()

        button.Bind(
            wx.EVT_BUTTON,
            lambda evt: wx.PostEvent(self.canvas, ToolChangeEvent(tool=name)),
        )



from __future__ import annotations

from typing import Dict, Union, Type, Iterable
import os


class ResourceItem:
    __slots__ = ("_path",)

    def __init__(self, path):
        self._path = path

    def __repr__(self):
        return f'{self.__class__.__name__}("{self._path}")'

    @property
    def path(self):
        return self._path

    def to(self, resource_item):
        return resource_item(self._path)

    @classmethod
    def default_object(cls):
        return None


class MissingResourceItem(ResourceItem):
    __slots__ = ("_path", "_factory")

    def __init__(self, path, factory):
        super(MissingResourceItem, self).__init__(path)
        self._factory = factory

    def __getattr__(self, item):
        return MissingResourceItem(os.path.join(self._path, item), self._factory)

    def __call__(self, *args, **kwargs):
        return self._factory.default_object()


class ResourceDirectory(ResourceItem):
    __slots__ = ("_path", "_factory", "_entries")
    ALLOWED_EXTS = ()

    def __init__(
        self, directory_name: str, factory_class: Type[ResourceItem], parent: str
    ):
        super(ResourceDirectory, self).__init__(os.path.join(parent, directory_name))
        self._factory = factory_class

        self._entries: Dict[str, Union[ResourceDirectory, ResourceItem, str]] = {}
        self.scan()

    def scan(self):
        self._entries.clear()
        for entry in os.listdir(self._path):
            path = os.path.join(self._path, entry)
            if os.path.isdir(path):
                self._entries[entry] = self.__class__(entry, self._factory, self._path)
            else:
                filename, ext = os.path.splitext(entry)
                ext = ext.lower()
                if ext in self.ALLOWED_EXTS:
                    self._entries[filename] = ext

    def __repr__(self):
        return f"ResourceDirectory({self._path})"

    def __getattr__(
        self, item
    ) -> Union["ResourceDirectory", ResourceItem, MissingResourceItem]:
        if item in self._entries:
            if isinstance(self._entries[item], str):
                ext = self._entries[item]
                self._entries[item] = self._factory(
                    os.path.join(self._path, f"{item}{ext}")
                )
            return self._entries[item]
        else:
            return MissingResourceItem(os.path.join(self._path, item), self._factory)

    def __dir__(self) -> Iterable[str]:
        return [k for k in self._entries.keys() if k != "__pycache__"]


# legacy support
from . import image as img

# class TextResourceItem(_ResourceItem):
#     def text(self):
#         with open(self._path) as fp:
#             return fp.read()
#
#     @classmethod
#     def default_object(cls):
#         return ""
#
#
# class JSONResourceItem(_ResourceItem):
#     def json(self):
#         with open(self._path) as fp:
#             return json.load(fp)
#
#     @classmethod
#     def default_object(cls):
#         return dict()



import wx
from typing import Set, Dict, Tuple

from .window_container import WindowContainer
from .key_config import KeyType, serialise_key, KeybindGroup

ActionIDType = str


_InputPressEventType = wx.NewEventType()
EVT_INPUT_PRESS = wx.PyEventBinder(_InputPressEventType)


class InputPressEvent(wx.PyEvent):
    """An action run by a button being pressed."""

    def __init__(self, action_id: ActionIDType):
        wx.PyEvent.__init__(self, eventType=_InputPressEventType)
        self._action_id = action_id

    @property
    def action_id(self) -> ActionIDType:
        """The string id of the action triggered by the keypress"""
        return self._action_id


_InputReleaseEventType = wx.NewEventType()
EVT_INPUT_RELEASE = wx.PyEventBinder(_InputReleaseEventType)


class InputReleaseEvent(wx.PyEvent):
    """An action run by a button being released."""

    def __init__(self, action_id: ActionIDType):
        wx.PyEvent.__init__(self, eventType=_InputReleaseEventType)
        self._action_id = action_id

    @property
    def action_id(self) -> ActionIDType:
        """The string id of the action triggered by the key release"""
        return self._action_id


_InputHeldEventType = wx.NewEventType()
EVT_INPUT_HELD = wx.PyEventBinder(_InputHeldEventType)


class InputHeldEvent(wx.PyEvent):
    """An action run by a button being held.
    This action will run frequently."""

    def __init__(self, action_ids: Set[ActionIDType]):
        wx.PyEvent.__init__(self, eventType=_InputHeldEventType)
        self._action_ids = action_ids

    @property
    def action_ids(self) -> Set[ActionIDType]:
        """The string ids of the actions that have been held."""
        return self._action_ids


class Action:
    """A private class for helping with key management for an action."""

    def __init__(self, trigger_key: KeyType, modifier_keys: Tuple[KeyType, ...]):
        self._required_keys = set(modifier_keys + (trigger_key,))
        self._trigger_key = trigger_key
        self._modifier_keys = set(modifier_keys)

    @property
    def required_keys(self) -> Set[KeyType]:
        """The keys that need to remain pressed for this action to stay active."""
        return self._required_keys

    @property
    def trigger_key(self) -> KeyType:
        """The key that needs pressing for this action to start."""
        return self._trigger_key

    @property
    def modifier_keys(self) -> Set[KeyType]:
        """Additional keys that must already be pressed for this action to start."""
        return self._modifier_keys


class ButtonInput(WindowContainer):
    """A class to detect and store user inputs."""

    def __init__(self, window: wx.Window):
        super().__init__(window)

        self._registered_actions: Dict[ActionIDType, Action] = {}

        self._pressed_keys: Set[KeyType] = set()

        # The actions that have been started but not stopped.
        self._continuous_actions: Set[ActionIDType] = set()

        # timer to deal with persistent actions
        self._input_timer = wx.Timer(self.window)

    def bind_events(self):
        """Set up all events required to run."""

        # key press actions
        self.window.Bind(wx.EVT_LEFT_DOWN, self._press)
        self.window.Bind(wx.EVT_LEFT_UP, self._release)
        self.window.Bind(wx.EVT_MIDDLE_DOWN, self._press)
        self.window.Bind(wx.EVT_MIDDLE_UP, self._release)
        self.window.Bind(wx.EVT_RIGHT_DOWN, self._press)
        self.window.Bind(wx.EVT_RIGHT_UP, self._release)
        self.window.Bind(wx.EVT_KEY_DOWN, self._press)
        self.window.Bind(wx.EVT_KEY_UP, self._release)
        self.window.Bind(wx.EVT_NAVIGATION_KEY, self._release)
        self.window.Bind(wx.EVT_NAVIGATION_KEY, self._press)
        self.window.Bind(wx.EVT_MOUSEWHEEL, self._release)
        self.window.Bind(wx.EVT_MOUSEWHEEL, self._press)
        self.window.Bind(wx.EVT_MOUSE_AUX1_DOWN, self._press)
        self.window.Bind(wx.EVT_MOUSE_AUX1_UP, self._release)
        self.window.Bind(wx.EVT_MOUSE_AUX2_DOWN, self._press)
        self.window.Bind(wx.EVT_MOUSE_AUX2_UP, self._release)

        self.window.Bind(
            wx.EVT_TIMER, self._process_continuous_inputs, self._input_timer
        )

        # save destruction
        self.window.Bind(wx.EVT_WINDOW_DESTROY, self._on_destroy, self.window)

    def enable(self):
        self._input_timer.Start(33)

    def disable(self):
        self._input_timer.Stop()

    def _on_destroy(self, evt):
        self.disable()
        evt.Skip()

    @property
    def pressed_keys(self) -> Set[KeyType]:
        """A tuple of all the keys that are currently pressed."""
        return self._pressed_keys.copy()

    @property
    def pressed_actions(self) -> Set[ActionIDType]:
        return self._continuous_actions.copy()

    def is_key_pressed(self, key: KeyType):
        """Is the requested key currently in the pressed state."""
        return key in self._pressed_keys

    def unpress_all(self):
        """Unpress all keys.
        This is useful if the window focus is lost because key release events will not be detected.
        """
        self._pressed_keys.clear()
        self._clean_up_actions()

    def clear_registered_actions(self):
        """Clear the previously registered actions so that they can be repopulated."""
        self._registered_actions.clear()
        self._continuous_actions.clear()

    def action_id_registered(self, action_id: ActionIDType):
        """Has the action id already been registered."""
        return action_id in self._registered_actions

    def register_action(
        self,
        action_id: ActionIDType,
        modifier_keys: Tuple[KeyType, ...],
        trigger_key: KeyType,
    ):
        """Register a new action for the given trigger key and optional modifier keys.
        This action will be fired when the trigger key is pressed providing all the modifier keys are already pressed.

        :param action_id: The unique action id. Will raise a ValueError if it is already taken.
        :param modifier_keys: Other keys that need to be pressed for the action to happen.
        :param trigger_key: The key that when pressed will start the action provided the modifier keys are pressed.
        :return:
        """
        if not isinstance(action_id, str):
            raise TypeError("action_id must be a string.")
        if (
            not isinstance(trigger_key, (str, int))
            or not isinstance(modifier_keys, tuple)
            or not all(isinstance(k, (str, int)) for k in modifier_keys)
        ):
            raise TypeError(
                "The key inputs are not of the correct format. Expected Union[str, int], Tuple[Union[str, int], ...]"
            )
        if self.action_id_registered(action_id):
            raise ValueError(f"{action_id} has already been registered.")
        self._registered_actions[action_id] = Action(trigger_key, modifier_keys)

    def register_actions(self, actions: KeybindGroup):
        for action_id, (modifier_keys, trigger_key) in actions.items():
            self.register_action(action_id, modifier_keys, trigger_key)

    def _find_actions(self, key: KeyType) -> Tuple[ActionIDType, ...]:
        """A method to find all actions triggered by `key` with the modifier keys also pressed."""
        return tuple(
            action_id
            for action_id, action in self._registered_actions.items()
            if action.trigger_key == key
            and action.modifier_keys.issubset(self._pressed_keys)
        )

    def _press(self, evt):
        """Event to handle a number of different key presses"""
        key = serialise_key(evt)
        if key is None:
            return
        if not self.is_key_pressed(key):
            action_ids = self._find_actions(key)
            self._continuous_actions.update(action_ids)
            for action_id in action_ids:
                wx.PostEvent(self.window, InputPressEvent(action_id))

            self._pressed_keys.add(key)
        evt.Skip()

    def _release(self, evt):
        """Event to handle a number of different key releases"""
        key = serialise_key(evt)
        if key is None:
            return
        if self.is_key_pressed(key):
            # remove the pressed key
            self._pressed_keys.remove(key)

        self._clean_up_actions()
        evt.Skip()

    def _clean_up_actions(self):
        # find all actions that are now not valid and remove them
        for action_id in list(self._continuous_actions):
            if not self._registered_actions[action_id].required_keys.issubset(
                self._pressed_keys
            ):
                self._continuous_actions.remove(action_id)
                wx.PostEvent(self.window, InputReleaseEvent(action_id))

    def _process_continuous_inputs(self, evt):
        wx.PostEvent(self.window, InputHeldEvent(self._continuous_actions.copy()))



class CanvasToggleElement:
    """This is a base class for any UI element that is dropped
    into the canvas and can be enabled and disabled."""

    def enable(self):
        """Set the state of the UI for being enabled.
        Do not bind events to the canvas here because they will get removed. Do this in bind_events.
        """
        raise NotImplementedError

    def bind_events(self):
        """Bind all required events to the canvas.
        All events on the canvas will be automatically removed after the UI is disabled.
        """
        raise NotImplementedError

    def disable(self):
        """Stop the UI. Unload any excessive data. May get resumed again with a call to enable.
        All events bound to the canvas will be automatically removed after this is run.
        """
        raise NotImplementedError



from .select_location import SelectLocationUI
from .events import (
    LocationChangeEvent,
    EVT_LOCATION_CHANGE,
    RotationChangeEvent,
    EVT_ROTATION_CHANGE,
    ScaleChangeEvent,
    EVT_SCALE_CHANGE,
    TransformChangeEvent,
    EVT_TRANSFORM_CHANGE,
)
from .select_transform import SelectTransformUI



import wx
from typing import TYPE_CHECKING

from amulet_map_editor.api import lang
from amulet_map_editor.api.framework import app
from amulet_map_editor.api.wx.ui.simple import SimplePanel
from amulet_map_editor.api.wx.ui.select_world import WorldUI
from amulet_map_editor.api.framework.programs import BaseProgram

if TYPE_CHECKING:
    from amulet.api.level import BaseLevel


class AboutProgram(SimplePanel, BaseProgram):
    def __init__(self, container, world: "BaseLevel"):
        SimplePanel.__init__(self, container)
        self.world = world

        self._close_world_button = wx.Button(
            self, wx.ID_ANY, label=lang.get("world.close_world")
        )
        self._close_world_button.Bind(wx.EVT_BUTTON, self._close_world)
        self.add_object(self._close_world_button, 0, wx.ALL | wx.CENTER)

        self.add_object(
            wx.StaticText(
                self,
                label=lang.get("program_about.currently_opened_world"),
            ),
            0,
            wx.ALL | wx.CENTER,
        )
        self.add_object(WorldUI(self, self.world.level_wrapper), 0, wx.ALL | wx.CENTER)
        self.add_object(
            wx.StaticText(
                self,
                label=f"{lang.get('program_about.choose_from_options')}\n<=================",
            ),
            0,
            wx.ALL | wx.CENTER,
        )

    def _close_world(self, evt):
        app.close_level(self.world.level_path)



from __future__ import annotations
import wx
from wx.lib.agw import flatnotebook
from typing import Dict, Union
import traceback
import logging

from amulet.api.errors import LoaderNoneMatched
from amulet_map_editor.api import config
from amulet_map_editor.api.wx.ui.select_world import open_level_from_dialog
from amulet_map_editor.api.wx.ui.traceback_dialog import TracebackDialog
from amulet_map_editor import __version__, lang
from amulet_map_editor.api.framework.pages import WorldPageUI
from .pages import AmuletMainMenu, BasePageUI
from .warning_dialog import WarningDialog

from amulet_map_editor.api import image
from amulet_map_editor.api.wx.util.ui_preferences import preserve_ui_preferences

log = logging.getLogger(__name__)

# Uses a conditional so if this breaks a build, we can just delete the file and it will skip the check
try:
    from amulet_map_editor.api.framework import update_check
except ImportError:
    update_check = None
    log.warning("Could not import update checker")

NOTEBOOK_MENU_STYLE = (
    flatnotebook.FNB_NO_X_BUTTON
    | flatnotebook.FNB_HIDE_ON_SINGLE_TAB
    | flatnotebook.FNB_NAV_BUTTONS_WHEN_NEEDED
)
NOTEBOOK_STYLE = NOTEBOOK_MENU_STYLE | flatnotebook.FNB_X_ON_TAB

CLOSEABLE_PAGE_TYPE = Union[WorldPageUI]

wx.Image.SetDefaultLoadFlags(0)


@preserve_ui_preferences
class AmuletUI(wx.Frame):
    """This is the top level frame that Amulet exists within."""

    # The notebook to hold world pages
    _level_notebook: AmuletLevelNotebook

    def __init__(self, parent):
        wx.Frame.__init__(
            self,
            parent,
            id=wx.ID_ANY,
            title=f"Amulet V{__version__}",
            pos=wx.DefaultPosition,
            size=wx.Size(1000, 600),
            style=wx.CAPTION
            | wx.CLOSE_BOX
            | wx.MINIMIZE_BOX
            | wx.MAXIMIZE_BOX
            | wx.SYSTEM_MENU
            | wx.TAB_TRAVERSAL
            | wx.CLIP_CHILDREN
            | wx.RESIZE_BORDER,
        )
        self.SetMinSize((625, 440))
        icon = wx.Icon()
        icon.CopyFromBitmap(image.logo.amulet_logo.bitmap())
        self.SetIcon(icon)

        self._level_notebook = AmuletLevelNotebook(self, agwStyle=NOTEBOOK_MENU_STYLE)
        self._level_notebook.init()

        self.Bind(wx.EVT_CLOSE, self._level_notebook.on_app_close)

        meta_config = config.get("amulet_meta", {})
        if not meta_config.get("do_not_show_warning_dialog", False):
            warning_dialog = WarningDialog(self)
            warning_dialog.ShowModal()
            if warning_dialog.do_not_show_again:
                meta_config["do_not_show_warning_dialog"] = True
                config.put("amulet_meta", meta_config)

        if update_check:
            self.Bind(
                update_check.EVT_UPDATE_CHECK,
                lambda evt: update_check.UpdateDialog(
                    self, __version__, evt.GetVersion()
                ).ShowModal(),
            )
            update_check.check_for_update(self, __version__)

    def open_level(self, path: str):
        """Open a level. You should use the method in the app."""
        self._level_notebook.open_level(path)

    def close_level(self, path: str):
        """Close a given level. You should use the method in the app."""
        self._level_notebook.close_level(path)

    def create_menu(self):
        """
        Create the UI menu.

        Adds the top level menu items then extends it from the active page
        """
        menu_dict = {}
        menu_dict.setdefault(lang.get("menu_bar.file.menu_name"), {}).setdefault(
            "system", {}
        ).setdefault(
            lang.get("menu_bar.file.open_world"),
            lambda evt: open_level_from_dialog(self),
        )
        # menu_dict.setdefault(lang.get('menu_bar.file.menu_name'), {}).setdefault('system', {}).setdefault('Create World', lambda: self.world.save())
        menu_dict.setdefault(lang.get("menu_bar.file.menu_name"), {}).setdefault(
            "exit", {}
        ).setdefault(lang.get("menu_bar.file.quit"), lambda evt: self.Close())
        menu_dict = self._level_notebook.extend_menu(menu_dict)
        menu_bar = wx.MenuBar()
        for menu_name, menu_data in menu_dict.items():
            menu = wx.Menu()
            separator = False
            for menu_section in menu_data.values():
                if separator:
                    menu.AppendSeparator()
                separator = True
                for menu_item_name, menu_item_options in menu_section.items():
                    callback = None
                    menu_item_description = None
                    wx_id = None
                    if callable(menu_item_options):
                        callback = menu_item_options
                    elif isinstance(menu_item_options, tuple):
                        if len(menu_item_options) >= 1:
                            callback = menu_item_options[0]
                        if len(menu_item_options) >= 2:
                            menu_item_description = menu_item_options[1]
                        if len(menu_item_options) >= 3:
                            wx_id = menu_item_options[2]
                    else:
                        continue

                    if not menu_item_description:
                        menu_item_description = ""
                    if not wx_id:
                        wx_id = wx.ID_ANY

                    menu_item: wx.MenuItem = menu.Append(
                        wx_id, menu_item_name, menu_item_description
                    )
                    self.Bind(wx.EVT_MENU, callback, menu_item)
            menu_bar.Append(menu, menu_name)
        self.SetMenuBar(menu_bar)


class AmuletLevelNotebook(flatnotebook.FlatNotebook):
    """A notebook to hold all world tabs."""

    # The main menu tab
    _main_menu: AmuletMainMenu

    # Storage of open world tabs for easy lookup
    _open_worlds: Dict[str, CLOSEABLE_PAGE_TYPE]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.Bind(flatnotebook.EVT_FLATNOTEBOOK_PAGE_CLOSING, self._on_page_closing)
        self.Bind(wx.EVT_NOTEBOOK_PAGE_CHANGING, self._page_changing, self)
        self.Bind(wx.EVT_NOTEBOOK_PAGE_CHANGED, self._page_changed, self)

        self._main_menu = AmuletMainMenu(self)
        self._open_worlds = {}

    def init(self):
        self._add_world_tab(self._main_menu, lang.get("main_menu.tab_name"))

    def open_level(self, path: str):
        """Open a world panel add it to the notebook"""
        if path in self._open_worlds:
            self.SetSelection(self.GetPageIndex(self._open_worlds[path]))
        else:
            try:
                world = WorldPageUI(self, path)
            except LoaderNoneMatched as e:
                log.error(f"Could not find a loader for this world.\n{e}")
                wx.MessageBox(f"{lang.get('select_world.no_loader_found')}\n{e}")
            except Exception as e:
                log.error(lang.get("select_world.loading_world_failed"), exc_info=True)
                dialog = TracebackDialog(
                    self,
                    lang.get("select_world.loading_world_failed"),
                    str(e),
                    traceback.format_exc(),
                )
                dialog.ShowModal()
                dialog.Destroy()
            else:
                self._open_worlds[path] = world
                self._add_world_tab(world, world.world_name)

    def _add_world_tab(self, page: BasePageUI, obj_name: str):
        """Add a tab and enable it."""
        self.AddPage(page, obj_name, True)

    def close_level(self, path: str):
        """Close a given world and remove it from the notebook"""
        if path in self._open_worlds:
            world = self._open_worlds[path]
            # note we don't remove it from the dictionary here
            # delete page starts the deletion but it can be vetoed
            # it is deleted from the dictionary in _on_page_closing
            self.DeletePage(self.GetPageIndex(world))

    def _on_page_closing(self, evt: flatnotebook.EVT_FLATNOTEBOOK_PAGE_CLOSING):
        """Handle the page closing."""
        page: CLOSEABLE_PAGE_TYPE = self.GetPage(evt.GetSelection())
        if page is not self._main_menu:
            if page.can_disable() and page.can_close():
                path = page.path
                page.disable()
                page.close()
                del self._open_worlds[path]
            else:
                evt.Veto()

    def _page_changing(self, evt: wx.BookCtrlEvent):
        if (
            evt.GetOldSelection() != wx.NOT_FOUND
            and not self.GetPage(evt.GetOldSelection()).can_disable()
        ):
            evt.Veto()

    def _page_changed(self, evt: wx.BookCtrlEvent):
        """Handle the page changing."""
        if evt.GetOldSelection() != evt.GetSelection():
            if evt.GetOldSelection() != wx.NOT_FOUND:
                self.GetPage(evt.GetOldSelection()).disable()

            if self.GetCurrentPage() is self._main_menu:
                self.SetAGWWindowStyleFlag(NOTEBOOK_MENU_STYLE)
            else:
                self.SetAGWWindowStyleFlag(NOTEBOOK_STYLE)

        if self.GetCurrentPage() is not None:
            self.GetCurrentPage().enable()

    def on_app_close(self, evt: wx.CloseEvent):
        for path, page in list(self._open_worlds.items()):
            self.close_level(path)
        if self.GetPageCount() > 1:
            wx.MessageBox(lang.get("app.world_still_used"))
        else:
            evt.Skip()

    def extend_menu(self, menu_dict: dict) -> dict:
        return self.GetCurrentPage().menu(menu_dict)



from typing import TYPE_CHECKING, Type, Any, Callable, Tuple
import wx
from OpenGL.GL import (
    glClear,
    GL_DEPTH_BUFFER_BIT,
)

from amulet.api.data_types import BlockCoordinates

from amulet_map_editor import lang
from amulet_map_editor.api.wx.ui.simple import SimpleScrollablePanel
from amulet_map_editor.api.wx.util.validators import IntValidator
from amulet_map_editor.api.opengl.camera import Projection, Camera
from amulet_map_editor.programs.edit.api.events import EVT_SELECTION_CHANGE
from amulet_map_editor.programs.edit.api.behaviour.inspect_block_behaviour import (
    InspectBlockBehaviour,
)
from amulet_map_editor.programs.edit.api.behaviour.block_selection_behaviour import (
    BlockSelectionBehaviour,
    EVT_RENDER_BOX_CHANGE,
    RenderBoxChangeEvent,
    EVT_RENDER_BOX_DISABLE_INPUTS,
    EVT_RENDER_BOX_ENABLE_INPUTS,
)
from amulet_map_editor.programs.edit.api.ui.tool import DefaultBaseToolUI
from amulet_map_editor.programs.edit.api.key_config import (
    KeybindGroup,
)
from amulet_map_editor.programs.edit.api.ui.nudge_button import NudgeButton

if TYPE_CHECKING:
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas


class BaseSelectionMoveButton(NudgeButton):
    def __init__(
        self,
        parent: wx.Window,
        camera: Camera,
        keybinds: KeybindGroup,
        label: str,
        tooltip: str,
        selection: BlockSelectionBehaviour,
    ):
        super().__init__(parent, camera, keybinds, label, tooltip)
        self._selection = selection


class Point1MoveButton(BaseSelectionMoveButton):
    def _move(self, offset: Tuple[int, int, int]):
        ox, oy, oz = offset
        (x, y, z), point2 = self._selection.active_block_positions
        self._selection.active_block_positions = (x + ox, y + oy, z + oz), point2


class Point2MoveButton(BaseSelectionMoveButton):
    def _move(self, offset: Tuple[int, int, int]):
        ox, oy, oz = offset
        point1, (x, y, z) = self._selection.active_block_positions
        self._selection.active_block_positions = point1, (x + ox, y + oy, z + oz)


class SelectionMoveButton(BaseSelectionMoveButton):
    def _move(self, offset: Tuple[int, int, int]):
        ox, oy, oz = offset
        (x1, y1, z1), (x2, y2, z2) = self._selection.active_block_positions
        self._selection.active_block_positions = (x1 + ox, y1 + oy, z1 + oz), (
            x2 + ox,
            y2 + oy,
            z2 + oz,
        )


class SelectTool(wx.BoxSizer, DefaultBaseToolUI):
    _x1: wx.SpinCtrl
    _y1: wx.SpinCtrl
    _z1: wx.SpinCtrl
    _x2: wx.SpinCtrl
    _y2: wx.SpinCtrl
    _z2: wx.SpinCtrl

    def __init__(self, canvas: "EditCanvas"):
        wx.BoxSizer.__init__(self, wx.HORIZONTAL)
        DefaultBaseToolUI.__init__(self, canvas)

        self._selection = BlockSelectionBehaviour(self.canvas)
        self._inspect_block = InspectBlockBehaviour(self.canvas, self._selection)

        self._button_panel = SimpleScrollablePanel(canvas)
        button_sizer = wx.BoxSizer(wx.VERTICAL)
        self._button_panel.SetSizer(button_sizer)

        def add_button(
            label: str, tooltip: str, action: Callable[[wx.PyEventBinder], None]
        ):
            button = wx.Button(self._button_panel, label=label)
            button.SetToolTip(tooltip)
            button_sizer.Add(button, 0, wx.ALL | wx.EXPAND, 5)
            button.Bind(wx.EVT_BUTTON, action)

        add_button(
            lang.get("program_3d_edit.select_tool.delete_button"),
            lang.get("program_3d_edit.select_tool.delete_button_tooltip"),
            lambda evt: self.canvas.delete(),
        )
        add_button(
            lang.get("program_3d_edit.select_tool.copy_button"),
            lang.get("program_3d_edit.select_tool.copy_button_tooltip"),
            lambda evt: self.canvas.copy(),
        )
        add_button(
            lang.get("program_3d_edit.select_tool.cut_button"),
            lang.get("program_3d_edit.select_tool.cut_button_tooltip"),
            lambda evt: self.canvas.cut(),
        )
        add_button(
            lang.get("program_3d_edit.select_tool.paste_button"),
            lang.get("program_3d_edit.select_tool.paste_button_tooltip"),
            lambda evt: self.canvas.paste_from_cache(),
        )

        self.Add(self._button_panel, 0, wx.ALIGN_CENTER_VERTICAL)

        self._x1 = self._add_spin_ctrl(
            lang.get("program_3d_edit.select_tool.scroll_point_x1"),
            lang.get("program_3d_edit.select_tool.scroll_point_x1_tooltip"),
            (160, 215, 145),
        )
        self._y1 = self._add_spin_ctrl(
            lang.get("program_3d_edit.select_tool.scroll_point_y1"),
            lang.get("program_3d_edit.select_tool.scroll_point_y1_tooltip"),
            (160, 215, 145),
        )
        self._z1 = self._add_spin_ctrl(
            lang.get("program_3d_edit.select_tool.scroll_point_z1"),
            lang.get("program_3d_edit.select_tool.scroll_point_z1_tooltip"),
            (160, 215, 145),
        )
        self._x2 = self._add_spin_ctrl(
            lang.get("program_3d_edit.select_tool.scroll_point_x2"),
            lang.get("program_3d_edit.select_tool.scroll_point_x2_tooltip"),
            (150, 150, 215),
        )
        self._y2 = self._add_spin_ctrl(
            lang.get("program_3d_edit.select_tool.scroll_point_y2"),
            lang.get("program_3d_edit.select_tool.scroll_point_y2_tooltip"),
            (150, 150, 215),
        )
        self._z2 = self._add_spin_ctrl(
            lang.get("program_3d_edit.select_tool.scroll_point_z2"),
            lang.get("program_3d_edit.select_tool.scroll_point_z2_tooltip"),
            (150, 150, 215),
        )

        self._box_size_selector_fstring = lang.get(
            "program_3d_edit.select_tool.box_size_selector_fstring"
        )
        try:
            box_size_fstring = self._box_size_selector_fstring.format(x=0, y=0, z=0)
        except:
            self._box_size_selector_fstring = "dx={x},dy={y},dz={z}"
            box_size_fstring = self._box_size_selector_fstring.format(x=0, y=0, z=0)
        self._box_size_selector_text = wx.StaticText(
            self._button_panel, label=box_size_fstring, style=wx.ALIGN_CENTER_HORIZONTAL
        )
        self._box_size_selector_text.SetToolTip(
            lang.get("program_3d_edit.select_tool.box_size_selector_tooltip")
        )
        button_sizer.Add(self._box_size_selector_text, 0, wx.ALL | wx.EXPAND, 5)

        self._box_volume_text = wx.StaticText(
            self._button_panel, label="0x0x0=0", style=wx.ALIGN_CENTER_HORIZONTAL
        )
        button_sizer.Add(self._box_volume_text, 0, wx.ALL | wx.EXPAND, 5)
        self._box_volume_text.SetToolTip(
            lang.get("program_3d_edit.select_tool.box_size_tooltip")
        )

        self._point1_move = Point1MoveButton(
            self._button_panel,
            self.canvas.camera,
            self.canvas.key_binds,
            lang.get("program_3d_edit.select_tool.button_point1"),
            lang.get("program_3d_edit.select_tool.button_point1_tooltip"),
            self._selection,
        )
        self._point1_move.SetBackgroundColour((160, 215, 145))
        self._point1_move.Disable()
        button_sizer.Add(self._point1_move, 0, wx.ALL | wx.EXPAND, 5)

        self._point2_move = Point2MoveButton(
            self._button_panel,
            self.canvas.camera,
            self.canvas.key_binds,
            lang.get("program_3d_edit.select_tool.button_point2"),
            lang.get("program_3d_edit.select_tool.button_point2_tooltip"),
            self._selection,
        )
        self._point2_move.SetBackgroundColour((150, 150, 215))
        self._point2_move.Disable()
        button_sizer.Add(self._point2_move, 0, wx.ALL | wx.EXPAND, 5)

        self._selection_move = SelectionMoveButton(
            self._button_panel,
            self.canvas.camera,
            self.canvas.key_binds,
            lang.get("program_3d_edit.select_tool.button_selection_box"),
            lang.get("program_3d_edit.select_tool.button_selection_box_tooltip"),
            self._selection,
        )
        self._selection_move.SetBackgroundColour((255, 255, 255))
        self._selection_move.Disable()
        button_sizer.Add(self._selection_move, 0, wx.ALL | wx.EXPAND, 5)

    @property
    def name(self) -> str:
        return "Select"

    def bind_events(self):
        super().bind_events()
        self.canvas.Bind(EVT_RENDER_BOX_CHANGE, self._box_renderer_change)
        self.canvas.Bind(EVT_RENDER_BOX_DISABLE_INPUTS, self._disable_inputs)
        self.canvas.Bind(EVT_RENDER_BOX_ENABLE_INPUTS, self._enable_inputs)
        self.canvas.Bind(EVT_SELECTION_CHANGE, self._on_selection_change)
        self._selection.bind_events()
        self._inspect_block.bind_events()

    def enable(self):
        super().enable()
        self._selection.enable()
        self._pull_selection()
        self._point1_move.enable()
        self._point2_move.enable()
        self._selection_move.enable()

    def disable(self):
        super().disable()
        self._point1_move.disable()
        self._point2_move.disable()
        self._selection_move.disable()

    def _add_spin_ctrl(
        self, label: str, tooltip: str, colour: Tuple[int, int, int]
    ) -> wx.SpinCtrl:
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        self._button_panel.GetSizer().Add(sizer, 0, wx.EXPAND)
        name_text = wx.StaticText(self._button_panel, label=label)
        sizer.Add(name_text, flag=wx.ALIGN_CENTER | wx.ALL, border=5)
        obj = wx.SpinCtrl(
            self._button_panel,
            style=wx.SP_ARROW_KEYS
            | wx.TE_PROCESS_ENTER
            | wx.ALIGN_CENTER_VERTICAL
            | wx.WANTS_CHARS,
            min=-30000000,
            max=30000000,
        )
        sizer.Add(obj, 1, flag=wx.CENTER | wx.TOP | wx.BOTTOM | wx.RIGHT, border=5)
        obj.Bind(wx.EVT_SPINCTRL, self._box_input_change)
        obj.SetValidator(IntValidator())
        obj.Disable()
        obj.SetToolTip(tooltip)
        obj.SetBackgroundColour(colour)
        return obj

    def _box_input_change(self, _):
        self._selection.active_block_positions = (
            (self._x1.GetValue(), self._y1.GetValue(), self._z1.GetValue()),
            (self._x2.GetValue(), self._y2.GetValue(), self._z2.GetValue()),
        )

    def _box_renderer_change(self, evt: RenderBoxChangeEvent):
        self._update_selection_inputs(*evt.points)
        evt.Skip()

    def _on_selection_change(self, evt):
        self._pull_selection()
        evt.Skip()

    def _pull_selection(self):
        self._update_selection_inputs(*self._selection.active_block_positions)

    def _update_selection_inputs(
        self, point1: BlockCoordinates, point2: BlockCoordinates
    ):
        (x1, y1, z1), (x2, y2, z2) = point1, point2
        self._x1.SetValue(x1)
        self._y1.SetValue(y1)
        self._z1.SetValue(z1)
        self._x2.SetValue(x2)
        self._y2.SetValue(y2)
        self._z2.SetValue(z2)
        xdim = int(abs(x2 - x1))
        ydim = int(abs(y2 - y1))
        zdim = int(abs(z2 - z1))
        self._box_size_selector_text.SetLabel(
            self._box_size_selector_fstring.format(
                x=xdim,
                y=ydim,
                z=zdim,
            )
        )
        self._box_volume_text.SetLabel(
            f"{xdim + 1}x{ydim + 1}x{zdim + 1}={(xdim + 1)*(ydim + 1)*(zdim + 1):,}"
        )
        self.Layout()

    def _enable_inputs(self, evt):
        self._set_scroll_state(True)
        self._point1_move.Enable()
        self._point2_move.Enable()
        self._selection_move.Enable()
        evt.Skip()

    def _disable_inputs(self, evt):
        self._set_scroll_state(False)
        self._point1_move.Disable()
        self._point2_move.Disable()
        self._selection_move.Disable()
        evt.Skip()

    def _set_scroll_state(self, state: bool):
        for scroll in (self._x1, self._y1, self._z1, self._x2, self._y2, self._z2):
            scroll.Enable(state)

    def _on_draw(self, evt):
        self.canvas.renderer.start_draw()
        if self.canvas.camera.projection_mode == Projection.PERSPECTIVE:
            self.canvas.renderer.draw_sky_box()
            glClear(GL_DEPTH_BUFFER_BIT)
        self.canvas.renderer.draw_level()
        self._selection.draw()
        self.canvas.renderer.end_draw()



from amulet_map_editor.api.datatypes import MenuData
from amulet_map_editor.api.framework.base_tab import BaseTab


class BasePageUI(BaseTab):
    def menu(self, menu: MenuData) -> MenuData:
        return menu



import wx
from typing import TYPE_CHECKING, Tuple, Union, Type
from OpenGL.GL import (
    glClear,
    GL_DEPTH_BUFFER_BIT,
)
import math
import numpy
import weakref

from amulet.api.data_types import PointCoordinates
from amulet.api.level import BaseLevel
from amulet.api.structure import structure_cache
from amulet.operations.paste import paste_iter
from amulet.utils.matrix import (
    rotation_matrix_xyz,
    decompose_transformation_matrix,
    scale_matrix,
    transform_matrix,
)

from amulet_map_editor import lang
from amulet_map_editor.api import image
from amulet_map_editor.api.wx.util.validators import IntValidator, FloatValidator
from amulet_map_editor.api.wx.ui.simple import SimpleScrollablePanel
from amulet_map_editor.api.opengl.camera import Projection, Camera
from amulet_map_editor.api.opengl.mesh.level import RenderLevel
from amulet_map_editor.programs.edit.api.key_config import (
    KeybindGroup,
)
from amulet_map_editor.programs.edit.api.operations import OperationSuccessful
from amulet_map_editor.programs.edit.api.ui.nudge_button import NudgeButton
from amulet_map_editor.programs.edit.api.ui.tool import DefaultBaseToolUI
from amulet_map_editor.programs.edit.api.behaviour import StaticSelectionBehaviour
from amulet_map_editor.programs.edit.api.behaviour.pointer_behaviour import (
    PointerBehaviour,
    EVT_POINT_CHANGE,
    PointChangeEvent,
)
from amulet_map_editor.programs.edit.api.events import (
    InputPressEvent,
    EVT_INPUT_PRESS,
)
from amulet_map_editor.programs.edit.api.key_config import ACT_BOX_CLICK

if TYPE_CHECKING:
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas

BottomLeftRight = wx.BOTTOM | wx.LEFT | wx.RIGHT
BottomLeftRightCentre = BottomLeftRight | wx.ALIGN_CENTER_HORIZONTAL
BottomLeftRightExpand = BottomLeftRight | wx.EXPAND


class TupleInput(wx.FlexGridSizer):
    WindowCls: Union[Type[wx.SpinCtrl], Type[wx.SpinCtrlDouble]] = None
    Validator: Type[wx.Validator] = None

    def __init__(
        self,
        parent: wx.Window,
        x_str: str,
        y_str: str,
        z_str: str,
        min_value: int = -30_000_000,
        max_value: int = 30_000_000,
        start_value: int = 0,
        style=wx.SP_ARROW_KEYS,
    ):
        super().__init__(3, 2, 5, 5)

        x_label = wx.StaticText(parent, label=x_str)
        self.Add(x_label, 0, wx.ALIGN_CENTER)
        self.x = self.WindowCls(
            parent, min=min_value, max=max_value, initial=start_value, style=style
        )
        self.x.SetValidator(self.Validator())
        self.Add(self.x, 0, wx.EXPAND)

        y_label = wx.StaticText(parent, label=y_str)
        self.Add(y_label, 0, wx.ALIGN_CENTER)
        self.y = self.WindowCls(
            parent, min=min_value, max=max_value, initial=start_value, style=style
        )
        self.y.SetValidator(self.Validator())
        self.Add(self.y, 0, wx.EXPAND)

        z_label = wx.StaticText(parent, label=z_str)
        self.Add(z_label, 0, wx.ALIGN_CENTER)
        self.z = self.WindowCls(
            parent, min=min_value, max=max_value, initial=start_value, style=style
        )
        self.z.SetValidator(self.Validator())
        self.Add(self.z, 0, wx.EXPAND)
        self.AddGrowableCol(1)

    @property
    def value(self):
        return self.x.GetValue(), self.y.GetValue(), self.z.GetValue()

    @value.setter
    def value(self, value):
        self.x.SetValue(value[0])
        self.y.SetValue(value[1])
        self.z.SetValue(value[2])


class TupleIntInput(TupleInput):
    WindowCls = wx.SpinCtrl
    Validator = IntValidator

    @property
    def value(self) -> Tuple[int, int, int]:
        return self.x.GetValue(), self.y.GetValue(), self.z.GetValue()

    @value.setter
    def value(self, value: Tuple[int, int, int]):
        self.x.SetValue(value[0])
        self.y.SetValue(value[1])
        self.z.SetValue(value[2])


class TupleFloatInput(TupleInput):
    WindowCls = wx.SpinCtrlDouble
    Validator = FloatValidator

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.x.Bind(wx.EVT_MOUSEWHEEL, self._on_wheel)
        self.y.Bind(wx.EVT_MOUSEWHEEL, self._on_wheel)
        self.z.Bind(wx.EVT_MOUSEWHEEL, self._on_wheel)
        self.x.SetIncrement(1)
        self.y.SetIncrement(1)
        self.z.SetIncrement(1)

    @property
    def value(self) -> Tuple[float, float, float]:
        return self.x.GetValue(), self.y.GetValue(), self.z.GetValue()

    @value.setter
    def value(self, value: Tuple[float, float, float]):
        self.x.SetValue(value[0])
        self.y.SetValue(value[1])
        self.z.SetValue(value[2])

    def _on_wheel(self, evt: wx.MouseEvent):
        """Add scroll wheel behaviour to the input."""
        rotation = evt.GetWheelRotation()
        ctrl = evt.GetEventObject()
        if rotation > 0:
            ctrl.SetValue(ctrl.GetValue() + ctrl.GetIncrement())
            self._changed(ctrl)
        elif rotation < 0:
            ctrl.SetValue(ctrl.GetValue() - ctrl.GetIncrement())
            self._changed(ctrl)

    @staticmethod
    def _changed(ctrl: wx.SpinCtrlDouble):
        """Create a changed event for the SpinCtrlDouble."""
        evt = wx.SpinDoubleEvent(wx.wxEVT_SPINCTRLDOUBLE, ctrl.GetId(), ctrl.GetValue())
        evt.SetEventObject(ctrl)
        wx.PostEvent(ctrl, evt)


class RotationTupleInput(TupleFloatInput):
    def __init__(
        self,
        parent: wx.Window,
        x_str: str,
        y_str: str,
        z_str: str,
        increment=90,
    ):
        super().__init__(
            parent, x_str, y_str, z_str, -180, 180, style=wx.SP_ARROW_KEYS | wx.SP_WRAP
        )
        self.increment = increment
        self.x.Bind(wx.EVT_SPINCTRLDOUBLE, self._on_change)
        self.y.Bind(wx.EVT_SPINCTRLDOUBLE, self._on_change)
        self.z.Bind(wx.EVT_SPINCTRLDOUBLE, self._on_change)

    @property
    def value(self) -> Tuple[float, float, float]:
        return self.x.GetValue(), self.y.GetValue(), self.z.GetValue()

    @value.setter
    def value(self, value: Tuple[float, float, float]):
        self.x.SetValue(round(value[0]))
        self.y.SetValue(round(value[1]))
        self.z.SetValue(round(value[2]))

    @property
    def rotation_radians(self) -> Tuple[float, float, float]:
        return (
            math.radians(self.x.GetValue()),
            math.radians(self.y.GetValue()),
            math.radians(self.z.GetValue()),
        )

    @property
    def increment(self) -> int:
        return self.x.GetIncrement()

    @increment.setter
    def increment(self, increment: int):
        assert isinstance(increment, int) and increment >= 1
        self.x.SetIncrement(increment)
        self.y.SetIncrement(increment)
        self.z.SetIncrement(increment)
        self._round_values()

    @staticmethod
    def _on_change(evt: wx.SpinDoubleEvent):
        ctrl = evt.GetEventObject()
        ctrl.SetValue(
            int(round(ctrl.GetValue() / ctrl.GetIncrement())) * ctrl.GetIncrement()
        )
        evt.SetValue(ctrl.GetValue())
        evt.Skip()

    def _round_value(self, ctrl: wx.SpinCtrlDouble):
        ctrl.SetValue(
            int(round(ctrl.GetValue() / ctrl.GetIncrement())) * ctrl.GetIncrement()
        )
        self._changed(ctrl)

    def _round_values(self):
        self._round_value(self.x)
        self._round_value(self.y)
        self._round_value(self.z)


class MoveButton(NudgeButton):
    def __init__(
        self,
        parent: wx.Window,
        camera: Camera,
        keybinds: KeybindGroup,
        label: str,
        tooltip: str,
        paste_tool: "PasteTool",
    ):
        super().__init__(parent, camera, keybinds, label, tooltip)
        self._paste_tool = weakref.ref(paste_tool)

    def _move(self, offset: Tuple[int, int, int]):
        ox, oy, oz = offset
        x, y, z = self._paste_tool().location
        self._paste_tool().location = x + ox, y + oy, z + oz


class PasteTool(wx.BoxSizer, DefaultBaseToolUI):
    def __init__(self, canvas: "EditCanvas"):
        wx.BoxSizer.__init__(self, wx.HORIZONTAL)
        DefaultBaseToolUI.__init__(self, canvas)

        self._selection = StaticSelectionBehaviour(self.canvas)
        self._cursor = PointerBehaviour(self.canvas)
        self._moving = False
        self._is_enabled = False

        self._paste_panel = SimpleScrollablePanel(canvas)
        self._paste_sizer = wx.BoxSizer(wx.VERTICAL)
        self._paste_panel.SetSizer(self._paste_sizer)
        self.Add(self._paste_panel, 0, wx.ALIGN_CENTER_VERTICAL)

        def add_line():
            """add a line to the UI"""
            line = wx.StaticLine(self._paste_panel)
            self._paste_sizer.Add(line, 0, wx.BOTTOM | wx.EXPAND, 5)

        def add_tick_box(name: str, state: bool = True):
            tick = wx.CheckBox(self._paste_panel, label=name)
            tick.SetValue(state)
            self._paste_sizer.Add(
                tick,
                flag=BottomLeftRight,
                border=5,
            )
            return tick

        def add_label(name: str):
            label = wx.StaticText(self._paste_panel, label=name)
            label.SetFont(
                wx.Font(
                    12,
                    wx.FONTFAMILY_DEFAULT,
                    wx.FONTSTYLE_NORMAL,
                    wx.FONTWEIGHT_NORMAL,
                    True,
                )
            )
            self._paste_sizer.Add(label, 0, BottomLeftRightCentre, 5)

        self._paste_sizer.AddSpacer(5)
        add_label(lang.get("program_3d_edit.paste_tool.location_label"))
        self._location = TupleIntInput(
            self._paste_panel,
            lang.get("program_3d_edit.paste_tool.location_x_label"),
            lang.get("program_3d_edit.paste_tool.location_y_label"),
            lang.get("program_3d_edit.paste_tool.location_z_label"),
        )
        self._location.x.SetToolTip(
            lang.get("program_3d_edit.paste_tool.location_x_tooltip")
        )
        self._location.y.SetToolTip(
            lang.get("program_3d_edit.paste_tool.location_y_tooltip")
        )
        self._location.z.SetToolTip(
            lang.get("program_3d_edit.paste_tool.location_z_tooltip")
        )
        self._paste_sizer.Add(
            self._location,
            flag=BottomLeftRightExpand,
            border=5,
        )

        self._move_button = MoveButton(
            self._paste_panel,
            self.canvas.camera,
            self.canvas.key_binds,
            lang.get("program_3d_edit.paste_tool.move_selection_label"),
            lang.get("program_3d_edit.paste_tool.move_selection_tooltip"),
            self,
        )
        self._paste_sizer.Add(
            self._move_button,
            flag=BottomLeftRightExpand,
            border=5,
        )

        add_line()

        add_label(lang.get("program_3d_edit.paste_tool.rotation_label"))
        self._free_rotation = wx.CheckBox(
            self._paste_panel,
            label=lang.get("program_3d_edit.paste_tool.free_rotation_label"),
        )
        self._free_rotation.SetToolTip(
            lang.get("program_3d_edit.paste_tool.free_rotation_tooltip")
        )
        self._paste_sizer.Add(
            self._free_rotation,
            flag=BottomLeftRight,
            border=5,
        )

        self._rotation = RotationTupleInput(
            self._paste_panel,
            lang.get("program_3d_edit.paste_tool.rotation_x_label"),
            lang.get("program_3d_edit.paste_tool.rotation_y_label"),
            lang.get("program_3d_edit.paste_tool.rotation_z_label"),
        )
        self._rotation.x.SetToolTip(
            lang.get("program_3d_edit.paste_tool.rotation_x_tooltip")
        )
        self._rotation.y.SetToolTip(
            lang.get("program_3d_edit.paste_tool.rotation_y_tooltip")
        )
        self._rotation.z.SetToolTip(
            lang.get("program_3d_edit.paste_tool.rotation_z_tooltip")
        )
        self._paste_sizer.Add(
            self._rotation,
            flag=BottomLeftRightExpand,
            border=5,
        )
        self._free_rotation.Bind(wx.EVT_CHECKBOX, self._on_free_rotation_change)

        rotate_sizer = wx.BoxSizer(wx.HORIZONTAL)
        self._paste_sizer.Add(
            rotate_sizer,
            flag=BottomLeftRightCentre,
            border=5,
        )

        self._rotate_left_button = wx.BitmapButton(
            self._paste_panel, bitmap=image.icon.tablericons.rotate_2.bitmap(22, 22)
        )
        self._rotate_left_button.SetToolTip(
            lang.get("program_3d_edit.paste_tool.rotate_anti_clockwise_tooltip")
        )
        self._rotate_left_button.Bind(wx.EVT_BUTTON, self._on_rotate_left)
        rotate_sizer.Add(self._rotate_left_button)

        self._rotate_right_button = wx.BitmapButton(
            self._paste_panel,
            bitmap=image.icon.tablericons.rotate_clockwise_2.bitmap(22, 22),
        )
        self._rotate_right_button.SetToolTip(
            lang.get("program_3d_edit.paste_tool.rotate_clockwise_tooltip")
        )
        self._rotate_right_button.Bind(wx.EVT_BUTTON, self._on_rotate_right)
        rotate_sizer.Add(self._rotate_right_button)

        add_line()

        add_label(lang.get("program_3d_edit.paste_tool.scale_label"))
        self._scale = TupleFloatInput(
            self._paste_panel,
            lang.get("program_3d_edit.paste_tool.scale_x_label"),
            lang.get("program_3d_edit.paste_tool.scale_y_label"),
            lang.get("program_3d_edit.paste_tool.scale_z_label"),
            start_value=1,
        )
        self._scale.x.SetToolTip(lang.get("program_3d_edit.paste_tool.scale_x_tooltip"))
        self._scale.y.SetToolTip(lang.get("program_3d_edit.paste_tool.scale_y_tooltip"))
        self._scale.z.SetToolTip(lang.get("program_3d_edit.paste_tool.scale_z_tooltip"))
        self._scale.x.SetDigits(2)
        self._scale.y.SetDigits(2)
        self._scale.z.SetDigits(2)
        self._paste_sizer.Add(
            self._scale,
            flag=BottomLeftRightExpand,
            border=5,
        )

        self._paste_panel.Bind(wx.EVT_SPINCTRL, self._on_transform_change)
        self._paste_panel.Bind(wx.EVT_SPINCTRLDOUBLE, self._on_transform_change)

        mirror_sizer = wx.BoxSizer(wx.HORIZONTAL)
        self._paste_sizer.Add(
            mirror_sizer,
            flag=BottomLeftRightCentre,
            border=5,
        )

        # the tablericons file names are the wrong way around
        self._mirror_horizontal_button = wx.BitmapButton(
            self._paste_panel,
            bitmap=image.icon.tablericons.flip_vertical.bitmap(22, 22),
        )
        self._mirror_horizontal_button.SetToolTip(
            lang.get("program_3d_edit.paste_tool.mirror_horizontal_tooltip")
        )
        self._mirror_horizontal_button.Bind(wx.EVT_BUTTON, self._on_mirror_horizontal)
        mirror_sizer.Add(self._mirror_horizontal_button)

        self._mirror_vertical_button = wx.BitmapButton(
            self._paste_panel,
            bitmap=image.icon.tablericons.flip_horizontal.bitmap(22, 22),
        )
        self._mirror_vertical_button.SetToolTip(
            lang.get("program_3d_edit.paste_tool.mirror_vertical_tooltip")
        )
        self._mirror_vertical_button.Bind(wx.EVT_BUTTON, self._on_mirror_vertical)
        mirror_sizer.Add(self._mirror_vertical_button)

        add_line()

        self._copy_air = add_tick_box(
            lang.get("program_3d_edit.paste_tool.copy_air_label")
        )
        self._copy_air.SetToolTip(
            lang.get("program_3d_edit.paste_tool.copy_air_tooltip")
        )
        self._copy_water = add_tick_box(
            lang.get("program_3d_edit.paste_tool.copy_water_label")
        )
        self._copy_water.SetToolTip(
            lang.get("program_3d_edit.paste_tool.copy_water_tooltip")
        )
        self._copy_lava = add_tick_box(
            lang.get("program_3d_edit.paste_tool.copy_lava_label")
        )
        self._copy_lava.SetToolTip(
            lang.get("program_3d_edit.paste_tool.copy_lava_tooltip")
        )

        add_line()

        confirm_button = wx.Button(self._paste_panel, label="Confirm")
        self._paste_sizer.Add(confirm_button, 0, BottomLeftRightExpand, 5)
        confirm_button.Bind(wx.EVT_BUTTON, self._paste_confirm)

        self._paste_panel.Disable()
        self.Layout()

    @property
    def name(self) -> str:
        return "Paste"

    def bind_events(self):
        super().bind_events()
        self._selection.bind_events()
        self._cursor.bind_events()
        self.canvas.Bind(EVT_POINT_CHANGE, self._on_pointer_change)
        self.canvas.Bind(EVT_INPUT_PRESS, self._on_input_press)

    def enable(self):
        super().enable()
        self._move_button.enable()
        self._selection.update_selection()
        self._moving = False

    def set_state(self, state):
        if (
            isinstance(state, dict)
            and isinstance(state.get("structure"), BaseLevel)
            and isinstance(state.get("dimension"), str)
        ):
            structure = state["structure"]
            dimension = state["dimension"]
        elif structure_cache:
            structure, dimension = structure_cache.get_structure()
        else:
            wx.MessageBox("A structure needs to be copied before one can be pasted.")
            return

        self._paste_panel.Enable()
        self._is_enabled = True
        self.canvas.renderer.fake_levels.clear()
        self.canvas.renderer.fake_levels.append(
            structure, dimension, (0, 0, 0), (1, 1, 1), (0, 0, 0)
        )
        self._moving = True

    def disable(self):
        super().disable()
        self._move_button.disable()
        self._paste_panel.Disable()
        self._is_enabled = False
        self.canvas.renderer.fake_levels.clear()

    @property
    def location(self) -> PointCoordinates:
        """The location as specified in the UI."""
        return self._location.value

    @location.setter
    def location(self, location: PointCoordinates):
        """Set the location value.
        Will update the UI and the renderer."""
        self._location.value = location
        self._update_transform()

    def _on_free_rotation_change(self, evt):
        if self._free_rotation.GetValue():
            self._rotation.increment = 1
        else:
            self._rotation.increment = 90

    def _on_rotate_left(self, evt):
        self._rotate(-90)

    def _on_rotate_right(self, evt):
        self._rotate(90)

    def _rotate(self, angle: int):
        """Rotate the floating selection by the angle based on the camera rotation."""
        angle = math.radians(angle)
        ry, rx = self.canvas.camera.rotation
        if rx < -45:
            rotation_change = rotation_matrix_xyz(0, angle, 0)
        elif -45 <= rx < 45:
            if -135 <= ry < -45:
                # east
                rotation_change = rotation_matrix_xyz(angle, 0, 0)
            elif -45 <= ry < 45:
                # south
                rotation_change = rotation_matrix_xyz(0, 0, angle)
            elif 45 <= ry < 135:
                # west
                rotation_change = rotation_matrix_xyz(-angle, 0, 0)
            else:
                # north
                rotation_change = rotation_matrix_xyz(0, 0, -angle)
        else:
            rotation_change = rotation_matrix_xyz(0, -angle, 0)

        self._rotation.value = numpy.rad2deg(
            decompose_transformation_matrix(
                numpy.matmul(
                    rotation_change, rotation_matrix_xyz(*self._rotation_radians())
                )
            )[1]
        )
        self._update_transform()

    def _rotation_radians(self) -> Tuple[float, float, float]:
        return tuple(math.radians(v) for v in self._rotation.value)

    def _on_mirror_vertical(self, evt):
        ry, rx = self.canvas.camera.rotation
        if -45 <= rx < 45:
            # looking north, east, south or west vertical mirror is always in y
            self._mirror(1)
        elif -135 <= ry < -45 or 45 <= ry < 135:
            # looking down or up facing east or west
            self._mirror(0)
        else:
            # looking down or up facing north or south
            self._mirror(2)

    def _on_mirror_horizontal(self, evt):
        ry, rx = self.canvas.camera.rotation
        if -135 <= ry < -45 or 45 <= ry < 135:
            # facing east or west
            self._mirror(2)
        else:
            # facing north or south
            self._mirror(0)

    def _mirror(self, axis: int):
        """Mirror the selection in the given axis.

        :param axis: The axis to scale in 0=x, 1=y, 2=z
        :return:
        """
        scale = [(-1, 1, 1), (1, -1, 1), (1, 1, -1)][axis]
        self._scale.value, rotation, _ = decompose_transformation_matrix(
            numpy.matmul(
                scale_matrix(*scale),
                transform_matrix(
                    self._scale.value, self._rotation_radians(), (0, 0, 0)
                ),
            )
        )
        self._rotation.value = numpy.rad2deg(rotation)
        self._update_transform()

    def _on_pointer_change(self, evt: PointChangeEvent):
        if self._is_enabled and self._moving:
            self.canvas.renderer.fake_levels.active_transform = (
                evt.point,
                self._scale.value,
                self._rotation_radians(),
            )
            self._location.value = evt.point
        evt.Skip()

    def _on_transform_change(self, evt):
        self._update_transform()
        evt.Skip()

    def _update_transform(self):
        """Update the renderer with the new values."""
        self.canvas.renderer.fake_levels.active_transform = (
            self._location.value,
            self._scale.value,
            self._rotation_radians(),
        )

    def _on_input_press(self, evt: InputPressEvent):
        if evt.action_id == ACT_BOX_CLICK:
            if self._is_enabled:
                self._moving = not self._moving
                if self._moving:
                    self.canvas.renderer.fake_levels.active_transform = (
                        self._location.value,
                        self._scale.value,
                        self._rotation_radians(),
                    )
        evt.Skip()

    def _paste_operation(self):
        if all(self._scale.value):
            fake_levels = self.canvas.renderer.fake_levels
            level_index: int = fake_levels.active_level_index
            if level_index is not None:
                render_level: RenderLevel = fake_levels.render_levels[level_index]
                yield from paste_iter(
                    self.canvas.world,
                    self.canvas.dimension,
                    render_level.level,
                    render_level.dimension,
                    self._location.value,
                    self._scale.value,
                    self._rotation.value,
                    self._copy_air.GetValue(),
                    self._copy_water.GetValue(),
                    self._copy_lava.GetValue(),
                )
        else:
            raise OperationSuccessful(
                lang.get("program_3d_edit.paste_tool.zero_scale_message")
            )

    def _paste_confirm(self, evt):
        self.canvas.run_operation(self._paste_operation)

    def _on_draw(self, evt):
        self.canvas.renderer.start_draw()
        if self.canvas.camera.projection_mode == Projection.PERSPECTIVE:
            self.canvas.renderer.draw_sky_box()
            glClear(GL_DEPTH_BUFFER_BIT)
        self.canvas.renderer.draw_level()
        self.canvas.renderer.draw_fake_levels()
        self._selection.draw()
        self.canvas.renderer.end_draw()



from typing import List, Tuple
from setuptools import setup, find_packages
from wheel.bdist_wheel import bdist_wheel
from Cython.Build import cythonize
import glob
import sys
import numpy
import subprocess
import logging
import re
import versioneer


first_party = {
    "amulet-core",
    "amulet-nbt",
    "pymctranslate",
    "minecraft-resource-pack",
}


def freeze_requirements(packages: List[str]) -> List[str]:
    # Pip install the requirements to find the newest compatible versions
    # This makes sure that the source versions are using the same dependencies as the compiled version.
    # This also makes sure that the source version is using the newest version of the dependency.
    if any("~=" in r and r.split("~=", 1)[0].lower() in first_party for r in packages):
        print("pip-install")
        try:
            # make sure pip is up to date
            subprocess.run([sys.executable, "-m", "pip", "install", "--upgrade", "pip"])
            # run pip install
            subprocess.run(
                [sys.executable, "-m", "pip", "install", *packages, "--upgrade"]
            )
            # run pip freeze
            installed = (
                subprocess.check_output(
                    [sys.executable, "-m", "pip", "freeze"], encoding="utf-8"
                )
                .strip()
                .split("\n")
            )
            requirements_map = {r.split("==")[0].lower(): r for r in installed}

            print(installed, requirements_map)
            for index, requirement in enumerate(packages):
                if "~=" in requirement:
                    lib = requirement.split("~=")[0].strip().lower()
                    if lib in first_party and lib in requirements_map:
                        packages[index] = requirements_map[lib]
            print(f"Modified packages to {packages}")
        except Exception as e:
            print("Failed to bake versions:", e)
    return packages


GCCPattern = re.compile(r"gcc version (?P<major>\d+)\.(?P<minor>\d+)")
ClangPattern = re.compile(r"clang(?:-|\s+version\s+)(?P<major>\d+)\.(?P<minor>\d+)")


def get_openmp_args() -> Tuple[List[str], List[str], List[str], List[str]]:
    # This has been lifted from here https://github.com/cython/cython/blob/606bd8cf235149c3be6876d0f5ae60032c8aab6c/runtests.py
    import sysconfig
    from distutils import ccompiler

    def get_openmp_args_for(arg) -> Tuple[List[str], List[str]]:
        """arg == 'CC' or 'CXX'"""
        cc = (
            sysconfig.get_config_var(arg) or ccompiler.get_default_compiler()
        ).split()[0]
        if cc == "msvc":
            # Microsoft Visual C
            return ["/openmp"], []
        elif cc:
            # Try GCC and Clang
            try:
                out = subprocess.check_output([cc, "-v"]).decode()
            except ChildProcessError:
                logging.exception(f"Could not resolve unknown compiler {cc}")
            else:
                gcc_match = GCCPattern.search(out)
                if gcc_match:
                    if (gcc_match.group("major"), gcc_match.group("minor")) >= (4, 2):
                        return ["-fopenmp"], ["-fopenmp"]
                    return [], []
                clang_match = ClangPattern.search(out)
                if clang_match:
                    # if (clang_match.group("major"), clang_match.group("minor")) >= (3, 7):
                    #     return ['-fopenmp'], ['-fopenmp']
                    return [], []
        # If all else fails disable openmp
        return [], []

    omp_ccargs, omp_clargs = get_openmp_args_for("CC")
    omp_cppcargs, omp_cpplargs = get_openmp_args_for("CXX")

    return omp_ccargs, omp_clargs, omp_cppcargs, omp_cpplargs


# build cython extensions
if next(glob.iglob("amulet_map_editor/**/*.pyx", recursive=True), None):
    # This throws an error if it does not match any files
    omp_ccargs, omp_clargs, omp_cppcargs, omp_cpplargs = get_openmp_args()
    ext = cythonize(
        "amulet_map_editor/**/*.pyx",
        aliases={
            "OPENMP_CCARGS": omp_ccargs,
            "OPENMP_CLARGS": omp_clargs,
            "OPENMP_CPPCARGS": omp_cppcargs,
            "OPENMP_CPPLARGS": omp_cpplargs,
        },
    )
else:
    ext = ()


cmdclass = versioneer.get_cmdclass()


# There might be a better way of doing this
# The extra argument needs to be defined in the sdist
# so that it doesn't error. It doesn't actually use it.
class SDist(cmdclass["sdist"]):
    user_options = cmdclass["sdist"].user_options + [
        ("find-libs=", None, ""),
    ]

    def initialize_options(self):
        super().initialize_options()
        self.find_libs = None


class BDistWheel(bdist_wheel):
    user_options = bdist_wheel.user_options + [
        (
            "find-libs=",
            None,
            "Find and fix the newest version of first party libraries. Only used internally.",
        ),
    ]

    def initialize_options(self):
        super().initialize_options()
        self.find_libs = None

    def finalize_options(self):
        if self.find_libs:
            self.distribution.install_requires = freeze_requirements(
                list(self.distribution.install_requires)
            )
        super().finalize_options()


cmdclass["sdist"] = SDist
cmdclass["bdist_wheel"] = BDistWheel


setup(
    version=versioneer.get_version(),
    cmdclass=cmdclass,
    include_dirs=[numpy.get_include()],
    packages=find_packages(),
    ext_modules=ext,
)



import numpy
from typing import Optional, Tuple, List
from enum import Enum
import wx
from wx import glcanvas
import math
from ..canvas_container import CanvasContainer
from ..data_types import CameraLocationType, CameraRotationType

from ..matrix import (
    rotation_matrix_yx,
    perspective_matrix,
    displacement_matrix,
    orthographic_matrix,
    TransformationMatrixType,
)


class Projection(Enum):
    TOP_DOWN = 0
    PERSPECTIVE = 1


_CameraMoveChangeEventType = wx.NewEventType()
EVT_CAMERA_MOVED = wx.PyEventBinder(_CameraMoveChangeEventType)


class CameraMovedEvent(wx.PyEvent):
    """Run when the camera has moved or rotated."""

    def __init__(
        self, camera_location: CameraLocationType, camera_rotation: CameraRotationType
    ):
        wx.PyEvent.__init__(self, eventType=_CameraMoveChangeEventType)
        self._location = camera_location
        self._rotation = camera_rotation

    @property
    def camera_location(self) -> CameraLocationType:
        """The location of the camera. (x, y, z)"""
        return self._location

    @property
    def camera_rotation(self) -> CameraRotationType:
        """The rotation of the camera. (yaw, pitch).
        This should behave the same as how Minecraft handles it.
        """
        return self._rotation


_ProjectionChangeEventType = wx.NewEventType()
EVT_PROJECTION_CHANGED = wx.PyEventBinder(_ProjectionChangeEventType)


class ProjectionChangedEvent(wx.PyEvent):
    """Run when the projection of the camera has changed."""

    def __init__(self, projection: Projection):
        wx.PyEvent.__init__(self, eventType=_ProjectionChangeEventType)
        self._projection = projection

    @property
    def projection(self) -> Projection:
        """The location of the camera. (x, y, z)"""
        return self._projection


class Camera(CanvasContainer):
    """A class to hold the state information of the camera."""

    _bounds: Tuple[Tuple[float, float, float], Tuple[float, float, float]]
    _location: CameraLocationType
    _rotation: CameraRotationType
    _projection_mode: Projection
    _fov: List[float]
    _clipping: List[Tuple[float, float]]
    _aspect_ratio: float
    _projection_matrix: Optional[TransformationMatrixType]
    _transformation_matrix: Optional[TransformationMatrixType]

    __slots__ = (
        "_bounds",
        "_location",
        "_rotation",
        "_projection_mode",
        "_fov",
        "_clipping",
        "_aspect_ratio",
        "_projection_matrix",
        "_transformation_matrix",
    )

    def __init__(self, canvas: glcanvas.GLCanvas):
        super().__init__(canvas)
        self._bounds = (
            (-1_000_000_000, -1_000_000_000, -1_000_000_000),
            (1_000_000_000, 1_000_000_000, 1_000_000_000),
        )
        self._location = (0.0, 0.0, 0.0)
        self._rotation = (0.0, 0.0)
        self._projection_mode = Projection.PERSPECTIVE
        self._fov = [100.0, 70.0]
        self._clipping = [
            (-(10**5), 10**5),
            (0.1, 10000.0),
        ]
        self._aspect_ratio = 4 / 3
        self._projection_matrix = None
        self._transformation_matrix = None

    def _reset_matrix(self):
        self._projection_matrix = None
        self._transformation_matrix = None

    def _notify_moved(self):
        """Post a CameraMovedEvent"""
        wx.PostEvent(self.canvas, CameraMovedEvent(self.location, self.rotation))

    @property
    def projection_mode(self) -> Projection:
        return self._projection_mode

    @projection_mode.setter
    def projection_mode(self, projection_mode: Projection):
        assert (
            projection_mode in Projection
        ), f"{projection_mode} is not a valid projection."
        if self._projection_mode != projection_mode:
            self._projection_mode = projection_mode
            self._reset_matrix()
            wx.PostEvent(self.canvas, ProjectionChangedEvent(self.projection_mode))

    @property
    def location(self) -> CameraLocationType:
        """The location of the camera. (x, y, z)"""
        return self._location

    @location.setter
    def location(self, camera_location: CameraLocationType):
        """Set the location of the camera. (x, y, z).
        Generates EVT_CAMERA_MOVED on the parent canvas."""
        if self.set_location(camera_location):
            self._notify_moved()

    def set_location(self, camera_location: CameraLocationType) -> bool:
        """Set the location of the camera. (x, y, z)."""
        assert (
            len(camera_location) == 3
        ), "camera_location must be an iterable of three floats."
        camera_location = tuple(
            min(max(float(c), c_min), c_max)
            for c, c_min, c_max in zip(camera_location, *self._bounds)
        )
        if camera_location != self._location:
            self._reset_matrix()
            self._location = camera_location
            return True
        return False

    @property
    def rotation(self) -> CameraRotationType:
        """The rotation of the camera. (yaw, pitch).
        This should behave the same as how Minecraft handles it.
        """
        return self._rotation

    @rotation.setter
    def rotation(self, camera_rotation: CameraRotationType):
        """Set the rotation of the camera. (yaw, pitch).
        yaw (-180 to 180), pitch (-90 to 90)
        This should behave the same as how Minecraft handles it.
        Generates EVT_CAMERA_MOVED on the parent canvas."""
        if self.set_rotation(camera_rotation):
            self._notify_moved()

    def set_rotation(self, camera_rotation: CameraRotationType) -> bool:
        """Set the rotation of the camera. (yaw, pitch).
        yaw (-180 to 180), pitch (-90 to 90)
        This should behave the same as how Minecraft handles it."""
        assert (
            len(camera_rotation) == 2
        ), "camera_rotation must be an iterable of two floats."
        ry, rx = map(float, camera_rotation)
        if not -180 <= ry < 180:
            ry %= 360
            if ry >= 180:
                ry -= 360
        if not -90 <= rx <= 90:
            rx = max(min(rx, 90), -90)
        camera_rotation = (ry, rx)
        if camera_rotation != self._rotation:
            self._reset_matrix()
            self._rotation = camera_rotation
            return True
        return False

    @property
    def location_rotation(self) -> Tuple[CameraLocationType, CameraRotationType]:
        """Get the camera location and rotation in one property."""
        return self.location, self.rotation

    @location_rotation.setter
    def location_rotation(
        self, location_rotation: Tuple[CameraLocationType, CameraRotationType]
    ):
        """Set the camera location and rotation in one property.
        Generates EVT_CAMERA_MOVED on the parent canvas."""
        location, rotation = location_rotation
        moved = self.set_location(location)
        rotated = self.set_rotation(rotation)
        if moved or rotated:
            self._notify_moved()

    def _set_fov(self, mode: Projection, fov: float):
        self._fov[mode.value] = float(fov)
        self._reset_matrix()

    @property
    def fov(self) -> float:
        """The field of view of the camera.
        The value will vary based on the projection mode."""
        return self._fov[self.projection_mode.value]

    @fov.setter
    def fov(self, fov: float):
        """Set the field of view of the camera.
        The value will vary based on the projection mode."""
        self._set_fov(self.projection_mode, fov)

    @property
    def perspective_fov(self) -> float:
        """The field of view of the camera in degrees when in perspective mode."""
        return self._fov[Projection.PERSPECTIVE.value]

    @perspective_fov.setter
    def perspective_fov(self, fov: float):
        """Set the field of view of the camera in degrees when in perspective mode."""
        self._set_fov(Projection.PERSPECTIVE, fov)

    @property
    def orthographic_fov(self) -> float:
        """The field of view of the camera when in orthographic mode."""
        return self._fov[Projection.TOP_DOWN.value]

    @orthographic_fov.setter
    def orthographic_fov(self, fov: float):
        """Set the field of view of the camera when in orthographic mode."""
        self._set_fov(Projection.TOP_DOWN, fov)

    def _set_clipping(self, mode: Projection, clipping: Tuple[float, float]):
        assert len(clipping) == 2, "camera_rotation must be an iterable of two floats."
        self._clipping[mode.value] = tuple(map(float, clipping))
        self._reset_matrix()

    @property
    def perspective_clipping(self) -> Tuple[float, float]:
        """The near and far clipping distance when in perspective mode."""
        return self._clipping[Projection.PERSPECTIVE.value]

    @perspective_clipping.setter
    def perspective_clipping(self, clipping: Tuple[float, float]):
        """Set the near and far clipping distance when in perspective mode."""
        self._set_clipping(Projection.PERSPECTIVE, clipping)

    @property
    def orthographic_clipping(self) -> Tuple[float, float]:
        """The near and far clipping distance when in orthographic mode."""
        return self._clipping[Projection.TOP_DOWN.value]

    @orthographic_clipping.setter
    def orthographic_clipping(self, clipping: Tuple[float, float]):
        """Set the near and far clipping distance when in orthographic mode."""
        self._set_clipping(Projection.TOP_DOWN, clipping)

    @property
    def aspect_ratio(self) -> float:
        """The aspect ratio of the camera (width/weight)"""
        return self._aspect_ratio

    @aspect_ratio.setter
    def aspect_ratio(self, aspect_ratio: float):
        """Set the aspect ratio of the camera (width/weight)"""
        self._aspect_ratio = float(aspect_ratio)
        self._reset_matrix()

    @staticmethod
    def rotation_matrix(yaw, pitch) -> TransformationMatrixType:
        """Helper function to get a rotation matrix for yaw and pitch.

        :param yaw: Yaw in degrees.
        :param pitch: Pitch in degrees.
        :return: A 4x4 rotation matrix.
        """
        return rotation_matrix_yx(math.radians(yaw + 180), math.radians(pitch))

    @property
    def camera_matrix(self) -> TransformationMatrixType:
        """The matrix to convert world space to camera space."""
        return numpy.matmul(
            self.rotation_matrix(*self.rotation),
            displacement_matrix(*-numpy.array(self.location)),
        )

    @property
    def projection_matrix(self) -> TransformationMatrixType:
        """The matrix to convert camera space to screen space."""
        if self._projection_matrix is None:
            if self.projection_mode == Projection.TOP_DOWN:
                self._projection_matrix = self.orthographic_matrix
            else:
                self._projection_matrix = self.perspective_matrix
            self._projection_matrix.flags.writeable = False

        return self._projection_matrix

    @property
    def orthographic_matrix(self) -> TransformationMatrixType:
        """The orthographic matrix to convert camera space to screen space."""
        near, far = self._clipping[self.projection_mode.value]
        return orthographic_matrix(self.fov, self.aspect_ratio, near, far)

    @property
    def perspective_matrix(self) -> TransformationMatrixType:
        """The perspective matrix to convert camera space to screen space."""
        z_near, z_far = self._clipping[self.projection_mode.value]
        return perspective_matrix(
            math.radians(self.fov), self.aspect_ratio, z_near, z_far
        )

    @property
    def transformation_matrix(self) -> TransformationMatrixType:
        """The world to projection matrix."""
        # camera translation
        if self._transformation_matrix is None:
            self._transformation_matrix = numpy.matmul(
                self.projection_matrix, self.camera_matrix
            )
            self._transformation_matrix.flags.writeable = False

        return self._transformation_matrix
